# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/live/util.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12478 bytes
from functools import singledispatch
from typing import Any, Union

import Live
from Live import Clip, ClipSlot, DeviceParameter, MixerDevice, Scene

from ableton.v2.base import liveobj_valid
from ableton.v2.control_surface.components import find_nearest_color
from ableton.v2.control_surface.internal_parameter import InternalParameterBase

from ..base import clamp, find_if, hex_to_rgb

TRANSLATED_MIXER_PARAMETER_NAMES = {"Track Volume": "Volume", "Track Panning": "Pan"}
UNDECLARED_QUANTIZED_PARAMETERS = {
    "AutoFilter": ("LFO Sync Rate",),
    "AutoPan": ("Sync Rate",),
    "BeatRepeat": ("Gate", "Grid", "Interval", "Offset", "Variation"),
    "Corpus": ("LFO Sync Rate",),
    "Flanger": ("Sync Rate",),
    "FrequencyShifter": ("Sync Rate",),
    "GlueCompressor": ("Ratio", "Attack", "Release"),
    "MidiArpeggiator": ("Offset", "Synced Rate", "Repeats", "Ret. Interval", "Transp. Steps"),
    "MidiNoteLength": ("Synced Length",),
    "MidiScale": ("Base",),
    "MultiSampler": ("L 1 Sync Rate", "L 2 Sync Rate", "L 3 Sync Rate"),
    "Operator": ("LFO Sync",),
    "OriginalSimpler": ("L Sync Rate",),
    "Phaser": ("LFO Sync Rate",),
}
_current_song = None


class LiveObjectTypeError(Exception):
    pass


def raise_type_error_for_liveobj(obj, return_value: Any = False) -> Any:
    return return_value


def application():
    return Live.Application.get_application()


def song():
    global _current_song
    if liveobj_valid(_current_song):
        return _current_song
    _current_song = application().get_document()
    return _current_song


def set_song(song_obj, from_test=False):
    global _current_song
    _current_song = song_obj


def get_bar_length(clip=None):
    obj = clip if liveobj_valid(clip) else song()
    return 4.0 * obj.signature_numerator / obj.signature_denominator


def is_arrangement_view_active():
    return application().view.focused_document_view == "Arranger"


def is_song_recording():
    return song().session_record or song().record_mode


def is_track_recording(track):
    playing_slot = playing_clip_slot(track)
    return liveobj_valid(playing_slot) and playing_slot.is_recording


def is_clip_new_recording(clip):
    return clip.is_recording and not clip.is_overdubbing


def playing_clip_slot(track):
    if track.clip_slots:
        index = track.playing_slot_index
        if index >= 0:
            return track.clip_slots[index]
        return None
    return None


def prepare_new_clip_slot(track, stop=False):
    slot = None
    if track.clip_slots:
        try:
            new_scene_index = _next_empty_clip_slot_index(track)
            song().view.selected_scene = song().scenes[new_scene_index]
            slot = track.clip_slots[scene_index()]
            if stop:
                track.stop_all_clips(False)
        except Live.Base.LimitationError:
            pass

        return slot
    return None


def _next_empty_clip_slot_index(track):
    current_index = scene_index()
    scene_count = len(song().scenes)
    while track.clip_slots[current_index].has_clip:
        current_index += 1
        if current_index == scene_count:
            song().create_scene(scene_count)

    return current_index


def is_track_armed(track):
    return liveobj_valid(track) and track.can_be_armed and (track.arm or track.implicit_arm)


def any_track_armed():
    return any(t.can_be_armed and t.arm for t in song().tracks)


def find_parent_track(liveobj):
    parent = liveobj.canonical_parent
    if isinstance(parent, Live.Track.Track):
        return parent
    return find_parent_track(parent)


def all_visible_tracks():
    return list(tuple(song().visible_tracks) + tuple(song().return_tracks) + (song().master_track,))


def track_index(track=None, track_list=None):
    track = track or song().view.selected_track
    track_list = track_list or list(song().tracks)
    if track in track_list:
        return track_list.index(track)
    return None


def scene_index(scene=None):
    scene = scene or song().view.selected_scene
    return list(song().scenes).index(scene)


@singledispatch
def display_name(obj: Union[(Scene, Clip, ClipSlot, DeviceParameter)], strip_space=True):
    return raise_type_error_for_liveobj(obj, return_value="")


@display_name.register
def _(scene: Scene, strip_space=True):
    return liveobj_name(scene, strip_space) or f"Scene {scene_index(scene) + 1}"


@display_name.register
def _(clip: Clip, strip_space=True):
    return liveobj_name(clip, strip_space) or "Clip"


@display_name.register
def _(clip_slot: ClipSlot, strip_space=True):
    if liveobj_valid(clip_slot.clip):
        return display_name(clip_slot.clip, strip_space)
    return f"Slot {list(clip_slot.canonical_parent.clip_slots).index(clip_slot) + 1}"


@display_name.register
def _(parameter: DeviceParameter, **_):
    if liveobj_valid(parameter):
        try:
            return parameter.display_name
        except AttributeError:
            return TRANSLATED_MIXER_PARAMETER_NAMES.get(parameter.name, parameter.name)

        return ""
    return None


@display_name.register
def _(parameter: InternalParameterBase, **_):
    return parameter.name


def liveobj_name(obj, strip_space=True):
    if liveobj_valid(obj):
        if strip_space:
            return obj.name.strip()
        return obj.name
    return ""


def is_device_rack(device):
    return liveobj_valid(device) and device.can_have_chains


def get_parameter_by_name(name, device):
    if liveobj_valid(device):
        return find_if(lambda p: p.original_name == name and liveobj_valid(p) and p.is_enabled, device.parameters)
    return None


def normalized_parameter_value(parameter, value=None):
    result = 0.0
    if liveobj_valid(parameter):
        value = value if value is not None else parameter.value
        param_range = parameter.max - parameter.min
        result = (value - parameter.min) / param_range
    return clamp(result, 0.0, 1.0)


def parameter_value_to_midi_value(parameter, max_value=128):
    return int(normalized_parameter_value(parameter) * (max_value - 1))


def parameter_owner(parameter):
    if isinstance(parameter.canonical_parent, MixerDevice):
        return parameter.canonical_parent.canonical_parent
    return parameter.canonical_parent


def is_parameter_quantized(parameter, device):
    is_quantized = False
    if liveobj_valid(parameter):
        device_class = getattr(device, "class_name", None)
        is_quantized = (parameter.is_quantized) or (
            (device_class in UNDECLARED_QUANTIZED_PARAMETERS)
            and (parameter.name in UNDECLARED_QUANTIZED_PARAMETERS[device_class])
        )
    return is_quantized


def liveobj_color_to_midi_rgb_values(obj, default_values=(0, 0, 0)):
    if liveobj_valid(obj):
        return tuple(i // 2 for i in hex_to_rgb(obj.color))
    return default_values


def liveobj_color_to_value_from_palette(obj, palette=None, fallback_table=None, default_value=0):
    if liveobj_valid(obj):
        try:
            return palette[obj.color]
        except (KeyError, IndexError):
            return find_nearest_color(fallback_table, obj.color)

        return default_value
    return None


def deduplicate_parameters(parameters):
    param_names = set()
    to_remove = object()
    for i, parameter in enumerate(parameters):
        if getattr(parameter, "original_name", None) in param_names:
            parameters[i] = to_remove
        if getattr(parameter, "original_name", None):
            param_names.add(parameter.original_name)

    without_duplicates = [param for param in parameters if param is not to_remove]
    padding_length = len(parameters) - len(without_duplicates)
    return without_duplicates + [None] * padding_length


def major_version():
    return Live.Application.get_application().get_major_version()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/control_list.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11888 bytes
from functools import partial

from future.moves.itertools import zip_longest
from past.utils import old_div

from ...base import clamp, find_if, first, flatten, is_matrix, mixin, old_hasattr, product, second
from .control import Connectable, Control
from .radio_button import RadioButtonControl

_DYNAMIC_CONTROL_COUNT = None


class ControlList(Control):
    DYNAMIC_CONTROL_COUNT = _DYNAMIC_CONTROL_COUNT

    class State(Control.State):
        def __init__(
            self,
            control=None,
            manager=None,
            unavailable_color=None,
            extra_args=None,
            extra_kws=None,
            *a,
            **k,
        ):
            super(ControlList.State, self).__init__(manager=manager, control=control)
            self._control_elements = None
            self._control_type = control.control_type
            self._controls = []
            self._dynamic_create = False
            self._unavailable_color = unavailable_color if unavailable_color is not None else "DefaultButton.Disabled"
            self._extra_args = a
            self._extra_kws = k
            self.control_count = control.control_count

        @property
        def control_elements(self):
            return self._control_elements

        @property
        def control_count(self):
            return len(self._controls)

        @control_count.setter
        def control_count():
            pass

        @property
        def unavailable_color(self):
            return self._unavailable_color

        @unavailable_color.setter
        def unavailable_color(self, value):
            self._unavailable_color = value
            control_elements = self._control_elements or []
            for control, element in zip_longest(self._controls, control_elements):
                if not control:
                    if element:
                        self._send_unavailable_color(element)

        def _create_controls(self, count):
            if count > len(self._controls):
                self._controls.extend([self._make_control(i) for i in range(len(self._controls), count)])
            else:
                if count < len(self._controls):
                    self._disconnect_controls(self._controls[count:])
                    self._controls = self._controls[:count]

        def _disconnect_controls(self, controls):
            for control in controls:
                control._get_state(self._manager).disconnect()
                control._clear_state(self._manager)

        def _make_control(self, index):
            control = (self._control_type)(*self._extra_args, **self._extra_kws)
            control._event_listeners = self._event_listeners
            control_state = control._get_state(self._manager)
            if not old_hasattr(control_state, "index"):
                control_state.index = index
            else:
                raise RuntimeError("Cannot set 'index' attribute. Attribute already set.")
            return control

        def set_control_element(self, control_elements):
            self._control_elements = control_elements
            if self._dynamic_create:
                if len(control_elements or []) != len(self._control_element or []):
                    self._create_controls(len(control_elements or []))
            self._update_controls()

        def _update_controls(self):
            control_elements = self._control_elements or []
            for control, element in zip_longest(self._controls, control_elements):
                if control:
                    control._get_state(self._manager).set_control_element(element)
                if element:
                    element.reset_state()
                    self._send_unavailable_color(element)

        def _send_unavailable_color(self, element):
            if old_hasattr(element, "set_light"):
                element.set_light(self._unavailable_color)

        def __getitem__(self, index):
            return self._controls[index]._get_state(self._manager)

        def _on_value(self, value, *a, **k):
            pass

        def _register_value_slot(self, manager, control):
            pass

    def __init__(self, control_type=None, control_count=_DYNAMIC_CONTROL_COUNT, *a, **k):
        (super().__init__)(a, extra_args=a, extra_kws=k, **k)
        self.control_type = control_type
        self.control_count = control_count


class RadioButtonGroup(ControlList, RadioButtonControl):
    class State(ControlList.State, Connectable):
        requires_listenable_connected_property = True

        def __init__(self, *a, **k):
            self._checked_index = -1
            (super(RadioButtonGroup.State, self).__init__)(*a, **k)

        @property
        def checked_index(self):
            return self._checked_index

        @checked_index.setter
        def checked_index(self, index):
            if index != -1:
                self[index].is_checked = True
            else:
                checked_control = find_if(lambda c: c.is_checked, self)
                if checked_control is not None:
                    checked_control.is_checked = False

        def connect_property(self, *a):
            (super(RadioButtonGroup.State, self).connect_property)(*a)
            self.checked_index = self.connected_property_value

        def on_connected_property_changed(self, value):
            self.checked_index = value

        def _create_controls(self, count):
            super(RadioButtonGroup.State, self)._create_controls(count)
            self.checked_index = clamp(self._checked_index, -1, count - 1)

        def _make_control(self, index):
            control = super(RadioButtonGroup.State, self)._make_control(index)
            control_state = control._get_state(self._manager)
            control_state._on_checked = partial(self._on_checked, control_state)
            control_state.is_checked = index == self._checked_index
            return control

        def _on_checked(self, checked_control):
            for control in self._controls:
                control = control._get_state(self._manager)
                control.is_checked = control == checked_control

            self._checked_index = checked_control.index
            self.connected_property_value = self._checked_index

    def __init__(self, *a, **k):
        (super().__init__)(RadioButtonControl, *a, **k)


_DYNAMIC_MATRIX_DIMENSIONS = (None, None)


class MatrixControl(ControlList):
    DYNAMIC_DIMENSIONS = _DYNAMIC_MATRIX_DIMENSIONS

    class State(ControlList.State):
        def __init__(self, control=None, manager=None, dimensions=None, *a, **k):
            (super(MatrixControl.State, self).__init__)(control, manager, *a, **k)
            self._dimensions = (None, None)
            if dimensions is not None:
                self.dimensions = dimensions

        @property
        def dimensions(self):
            return self._dimensions

        @dimensions.setter
        def dimensions(self, dimensions):
            self._dynamic_create = dimensions == MatrixControl.DYNAMIC_DIMENSIONS
            if self._dynamic_create:
                count = len(self._control_elements) if self._control_elements else 0
            self._dimensions = dimensions
            count = first(dimensions) * second(dimensions)
            self._create_controls(count)
            self._update_controls()

        def _create_controls(self, count):
            super(MatrixControl.State, self)._create_controls(count)
            self._update_coordinates()

        def _make_control(self, index):
            control = super(MatrixControl.State, self)._make_control(index)
            if old_hasattr(control._get_state(self._manager), "coordinate"):
                raise RuntimeError("Cannot set 'coordinate' attribute. Attribute already set.")
            return control

        def _update_coordinates(self):
            for index, control in enumerate(self._controls):
                control_state = control._get_state(self._manager)
                control_state.coordinate = (int(old_div(index, self.width)), index % self.width)

        def set_control_element(self, control_elements):
            dimensions = (None, None)
            if old_hasattr(control_elements, "width") and old_hasattr(control_elements, "height"):
                dimensions = (control_elements.height(), control_elements.width())
                if not self._dynamic_create:
                    control_elements = [
                        control_elements.get_button(row, col)
                        for row, col in product(range(self.height), range(self.width))
                    ]
            else:
                if is_matrix(control_elements):
                    dimensions = (len(control_elements), len(first(control_elements)))
                    if not self._dynamic_create:
                        control_elements = [row[0 : self.width] for row in control_elements]
                    control_elements = list(flatten(control_elements))
                else:
                    if control_elements is not None:
                        raise RuntimeError("Control Elements must be a matrix")
            if self._dynamic_create:
                if None not in dimensions:
                    self._dimensions = dimensions
                    self._create_controls(first(dimensions) * second(dimensions))
                    self._update_controls()
            super(MatrixControl.State, self).set_control_element(control_elements)

        def get_control(self, row, column):
            index = row * self.width + column
            return self._controls[index]._get_state(self._manager)

        @property
        def width(self):
            return second(self._dimensions)

        @property
        def height(self):
            return first(self._dimensions)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)


_control_list_classes = {}
_control_matrix_classes = {}


def control_list(control_type, *a, **k):
    if control_type == RadioButtonControl:
        return RadioButtonGroup(*a, **k)
    c = _control_list_classes.get(control_type)
    if not c:
        c = mixin(ControlList, control_type)
        _control_list_classes[control_type] = c
    return c(control_type, *a, **k)


def control_matrix(control_type, *a, **k):
    m = _control_matrix_classes.get(control_type)
    if not m:
        m = mixin(MatrixControl, control_type)
        _control_matrix_classes[control_type] = m
    return m(control_type, *a, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SubjectSlot.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13732 bytes
import contextlib
from functools import partial, wraps
from itertools import repeat

from future.utils import with_metaclass
from past.builtins import basestring

from .Disconnectable import CompoundDisconnectable, Disconnectable
from .Signal import Signal
from .Util import NamedTuple, instance_decorator, mixin, monkeypatch, monkeypatch_extend


class SubjectSlotError(Exception):
    pass


class SubjectEvent(NamedTuple):
    name = None
    doc = ""
    signal = Signal
    override = False


def subject_add_event(cls, event_name_or_event):
    if isinstance(event_name_or_event, basestring):
        event = SubjectEvent(name=event_name_or_event)
    else:
        event = event_name_or_event
    signal_attr = "_" + event.name + "_signal"

    def get_signal(self):
        try:
            return getattr(self, signal_attr)
        except AttributeError:
            signal = event.signal(sender=self)
            setattr(self, signal_attr, signal)
            return signal

    kwargs = {"doc": event.doc, "override": event.override}

    @monkeypatch(cls, (event.name + "_has_listener"), **kwargs)
    def has_method(self, slot):
        return get_signal(self).is_connected(slot)

    @monkeypatch(cls, ("add_" + event.name + "_listener"), **kwargs)
    def add_method(self, slot, identify_sender=False, *a, **k):
        sender = self if identify_sender else None
        return (get_signal(self).connect)(slot, *a, sender=sender, **k)

    @monkeypatch(cls, ("remove_" + event.name + "_listener"), **kwargs)
    def remove_method(self, slot):
        return get_signal(self).disconnect(slot)

    @monkeypatch(cls, ("notify_" + event.name), **kwargs)
    def notify_method(self, *a, **k):
        return (get_signal(self))(*a, **k)

    @monkeypatch(cls, ("clear_" + event.name + "_listeners"), **kwargs)
    def clear_method(self):
        return get_signal(self).disconnect_all()

    @monkeypatch(cls, (event.name + "_listener_count"), **kwargs)
    def listener_count_method(self):
        return get_signal(self).count

    @monkeypatch_extend(cls)
    def disconnect(self):
        get_signal(self).disconnect_all()


def setup_subject(cls, listeners):
    for lst in listeners:
        subject_add_event(cls, lst)


class SubjectMeta(type):
    def __new__(cls, name, bases, dct):
        events = dct.get("__subject_events__", [])
        if events:
            if "disconnect" not in dct:
                dct["disconnect"] = lambda self: super(cls, self).disconnect()
        cls = super().__new__(cls, name, bases, dct)
        setup_subject(cls, events)
        return cls


class Subject(with_metaclass(SubjectMeta, Disconnectable)):
    pass


class SlotManager(CompoundDisconnectable):
    def register_slot(self, *a, **k):
        if a:
            slot = a[0] if isinstance(a[0], SubjectSlot) else SubjectSlot(*a, **k)
            self.register_disconnectable(slot)
            return slot
        return None

    def register_slot_manager(self, *a, **k):
        if a:
            manager = a[0] if isinstance(a[0], SlotManager) else SlotManager(*a, **k)
            self.register_disconnectable(manager)
            return manager
        return None


class SubjectSlot(Disconnectable):
    _extra_kws = {}
    _extra_args = []

    def __init__(self, subject=None, listener=None, event=None, extra_kws=None, extra_args=None, *a, **k):
        (super().__init__)(*a, **k)
        self._event = event
        if extra_kws is not None:
            self._extra_kws = extra_kws
        if extra_args is not None:
            self._extra_args = extra_args
        self._subject = None
        self._listener = None
        self.subject = subject
        self.listener = listener

    def disconnect(self):
        self.subject = None
        self.listener = None
        super().disconnect()

    def _check_subject_interface(self, subject):
        if not callable(getattr(subject, "add_" + self._event + "_listener", None)):
            raise SubjectSlotError(f'Subject {subject} missing "add" method for event: {self._event}')
        if not callable(getattr(subject, "remove_" + self._event + "_listener", None)):
            raise SubjectSlotError(f'Subject {subject} missing "remove" method for event: {self._event}')
        if not callable(getattr(subject, self._event + "_has_listener", None)):
            raise SubjectSlotError(f'Subject {subject} missing "has" method for event: {self._event}')

    def connect(self):
        if (self.is_connected or self._subject) is None or self._listener is not None:
            add_method = getattr(self._subject, "add_" + self._event + "_listener")
            all_args = (*tuple(self._extra_args), self._listener)
            with contextlib.suppress(RuntimeError):
                add_method(*all_args, **self._extra_kws)

    def soft_disconnect(self):
        if not self.is_connected and self._subject is not None or self._listener is not None:
            all_args = (*tuple(self._extra_args), self._listener)
            remove_method = getattr(self._subject, "remove_" + self._event + "_listener")
            with contextlib.suppress(RuntimeError):
                remove_method(*all_args)

    @property
    def is_connected(self):
        all_args = (*tuple(self._extra_args), self._listener)
        with contextlib.suppress(RuntimeError):
            return bool(
                self._subject is not None
                and self._listener is not None
                and (getattr(self._subject, self._event + "_has_listener"))(*all_args),
            )

    def _get_subject(self):
        return self._subject

    def _set_subject(self, subject):
        if subject != self._subject:
            if subject is not None:
                self._check_subject_interface(subject)
            self.soft_disconnect()
            self._subject = subject
            self.connect()

    subject = property(_get_subject, _set_subject)

    def _get_listener(self):
        return self._listener

    def _set_listener(self, listener):
        if listener != self._listener:
            self.soft_disconnect()
            self._listener = listener
            self.connect()

    listener = property(_get_listener, _set_listener)


class CallableSlotMixin:
    def __init__(self, function=None, *a, **k):
        (super().__init__)(*a, **k)
        self.function = function

    def __call__(self, *a, **k):
        return (self.function)(*a, **k)


class SubjectSlotGroup(SlotManager):
    listener = None
    _extra_kws = None
    _extra_args = None

    def __init__(self, listener=None, event=None, extra_kws=None, extra_args=None, *a, **k):
        (super().__init__)(*a, **k)
        self.listener = listener
        self._event = event
        if listener is not None:
            self.listener = listener
        if extra_kws is not None:
            self._extra_kws = extra_kws
        if extra_args is not None:
            self._extra_args = extra_args

    def replace_subjects(self, subjects, identifiers=repeat(None)):
        self.disconnect()
        for subject, identifier in zip(subjects, identifiers):
            self.add_subject(subject, identifier=identifier)

    def add_subject(self, subject, identifier=None):
        if identifier is None:
            identifier = subject
        listener = self._listener_for_subject(identifier)
        self.register_slot(subject, listener, self._event, self._extra_kws, self._extra_args)

    def remove_subject(self, subject):
        slot = self.find_disconnectable(
            lambda x: x.subject == subject,
        )
        self.disconnect_disconnectable(slot)

    def has_subject(self, subject):
        return (
            self.find_disconnectable(
                lambda x: x.subject == subject,
            )
            is not None
        )

    def _listener_for_subject(self, identifier):
        return lambda *a, **k: self.listener and (self.listener)(*(*a, identifier), **k)


class MultiSubjectSlot(SlotManager, SubjectSlot):
    def __init__(self, subject=None, listener=None, event=None, extra_kws=None, extra_args=None, *a, **k):
        self._original_listener = listener
        self._slot_subject = None
        self._nested_slot = None
        super().__init__(
            event=(event[0]),
            listener=(self._event_fired),
            subject=subject,
            extra_kws=extra_kws,
            extra_args=extra_args,
        )
        if len(event) > 1:
            self._nested_slot = self.register_disconnectable(
                MultiSubjectSlot(
                    event=(event[1:]),
                    listener=listener,
                    subject=subject,
                    extra_kws=extra_kws,
                    extra_args=extra_args,
                ),
            )
            self._update_nested_subject()

    def _get_subject(self):
        return super()._get_subject()

    def _set_subject(self, subject):
        try:
            try:
                super()._set_subject(subject)
            except SubjectSlotError:
                if self._nested_slot is None:
                    raise

        finally:
            self._slot_subject = subject
            self._update_nested_subject()

    subject = property(_get_subject, _set_subject)

    def _event_fired(self, *a, **k):
        self._update_nested_subject()
        (self._original_listener)(*a, **k)

    def _update_nested_subject(self):
        if self._nested_slot is not None:
            self._nested_slot.subject = (
                getattr(self._slot_subject, self._event) if self._slot_subject is not None else None
            )


def subject_slot(events, *a, **k):
    @instance_decorator
    def decorator(self, method):
        function = partial(method, self)
        event_list = events.split(".")
        num_events = len(event_list)
        event = event_list if num_events > 1 else events
        base_class = MultiSubjectSlot if num_events > 1 else SubjectSlot
        slot = wraps(method)(
            mixin(base_class, CallableSlotMixin)(
                event=event,
                extra_kws=k,
                extra_args=a,
                listener=function,
                function=function,
            ),
        )
        self.register_slot(slot)
        return slot

    return decorator


class CallableSubjectSlotGroup(SubjectSlotGroup, CallableSlotMixin):
    pass


def subject_slot_group(event, *a, **k):
    @instance_decorator
    def decorator(self, method):
        function = partial(method, self)
        slot = wraps(method)(
            CallableSubjectSlotGroup(event=event, extra_kws=k, extra_args=a, listener=function, function=function),
        )
        self.register_slot_manager(slot)
        return slot

    return decorator



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_UserScript/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13024 bytes
import logging
import os
from ast import parse
from configparser import ConfigParser
from functools import partial

from _Generic.GenericScript import TRANSPORT_BUTTON_SPECIFICATIONS, GenericScript
from ableton.v3.control_surface import MapMode

from .DeviceComponent import DeviceComponent

map_modes = {k: v for k, v in vars(MapMode).items() if k[:1] != "_" if not callable(v)}
logger = logging.getLogger(__name__)
HIDE_SCRIPT = True


def get_name_for_script(script_path):
    folder_name = str(script_path.split(os.path.sep)[-2].replace(" ", "_"))
    try:
        parse(f"{folder_name}= None")
        script_name = folder_name
    except SyntaxError:
        script_name = "GenericScript"

    return script_name


def parse_map_mode(config_parser, section_name, option_name):
    if config_parser.has_option(section_name, option_name):
        return map_modes.get(config_parser.get(section_name, option_name), "Absolute")
    return map_modes["Absolute"]


def parse_int(config_parser, integer_range, section_name, option_name):
    integer = -1
    if config_parser.has_option(section_name, option_name):
        try:
            option_value = config_parser.getint(section_name, option_name)
            if option_value in range(integer_range):
                integer = option_value
        except ValueError as e:
            try:
                logger.error("Error while parsing integer value %s", str(e))
            finally:
                e = None
                del e

        return integer
    return None


open_file = open


def create_instance(c_instance, user_path=""):
    device_map_mode = map_modes["Absolute"]
    volume_map_mode = map_modes["Absolute"]
    if user_path != "":
        file_object = open_file(user_path)
        if file_object:
            config_parser = ConfigParser()
            config_parser.read_file(file_object)
            parse_identifier = partial(parse_int, config_parser, 128)
            parse_channel = partial(parse_int, config_parser, 16)
            device_controls = [(-1, -1)] * 16
            transport_controls = {key.upper(): -1 for key in TRANSPORT_BUTTON_SPECIFICATIONS}
            volume_controls = [(-1, -1)] * 8
            trackarm_controls = [-1] * 8
            bank_controls = {
                "ONOFF": -1,
                "TOGGLELOCK": -1,
                "NEXTBANK": -1,
                "PREVBANK": -1,
                "BANK1": -1,
                "BANK2": -1,
                "BANK3": -1,
                "BANK4": -1,
                "BANK5": -1,
                "BANK6": -1,
                "BANK7": -1,
                "BANK8": -1,
            }
            controller_descriptions = {"CHANNEL": -1}
            mixer_options = {
                "NUMSENDS": 2,
                "INVERTMUTELEDS": True,
                "SEND1": [-1] * 8,
                "SEND2": [-1] * 8,
                "MUTE": [-1] * 8,
                "SOLO": [-1] * 8,
                "SELECT": [-1] * 8,
                "SENDMAPMODE": map_modes["Absolute"],
                "MASTERVOLUME": -1,
                "MASTERVOLUMECHANNEL": -1,
                "CUEVOLUME": -1,
                "CUEVOLUMECHANNEL": -1,
                "CROSSFADER": -1,
                "CROSSFADERCHANNEL": -1,
                "CROSSFADERMAPMODE": map_modes["Absolute"],
            }
            if config_parser.has_section("DeviceControls"):
                device_controls = [
                    (
                        parse_identifier("DeviceControls", f"Encoder{index + 1}"),
                        parse_channel("DeviceControls", f"EncoderChannel{index + 1}"),
                    )
                    for index in range(16)
                ]
                bank_controls["NEXTBANK"] = parse_identifier("DeviceControls", "NextBankButton")
                bank_controls["PREVBANK"] = parse_identifier("DeviceControls", "PrevBankButton")
                bank_controls["TOGGLELOCK"] = parse_identifier("DeviceControls", "LockButton")
                bank_controls["ONOFF"] = parse_identifier("DeviceControls", "OnOffButton")
                device_map_mode = parse_map_mode(config_parser, "DeviceControls", "EncoderMapMode")
            if config_parser.has_section("TransportControls"):
                for key in TRANSPORT_BUTTON_SPECIFICATIONS:
                    transport_controls[key.upper()] = parse_identifier("TransportControls", f"{key}Button")

            if config_parser.has_section("Globals"):
                controller_descriptions["CHANNEL"] = parse_channel("Globals", "GlobalChannel")
                pad_translation = []
                for pad in range(16):
                    pad_info = []
                    note = -1
                    channel = -1
                    option_name = f"Pad{pad + 1}Note"
                    if config_parser.has_option("Globals", option_name):
                        note = config_parser.getint("Globals", option_name)
                        if note in range(128):
                            option_name = f"Pad{pad + 1}Channel"
                            if config_parser.has_option("Globals", option_name):
                                channel = config_parser.getint("Globals", option_name)
                            if channel == -1:
                                if controller_descriptions["CHANNEL"] != -1:
                                    channel = controller_descriptions["CHANNEL"]
                            if channel in range(16):
                                pad_info.append(pad % 4)
                                pad_info.append(pad // 4)
                                pad_info.append(note)
                                pad_info.append(channel)
                                pad_translation.append(tuple(pad_info))

                if len(pad_translation) > 0:
                    controller_descriptions["PAD_TRANSLATION"] = tuple(pad_translation)
            if config_parser.has_section("MixerControls"):
                mixer_options["MASTERVOLUME"] = parse_identifier("MixerControls", "MasterVolumeSlider")
                mixer_options["MASTERVOLUMECHANNEL"] = parse_channel("MixerControls", "MasterSliderChannel")
                mixer_options["CUEVOLUME"] = parse_identifier("MixerControls", "CueVolumeSlider")
                mixer_options["CUEVOLUMECHANNEL"] = parse_channel("MixerControls", "CueSliderChannel")
                mixer_options["CROSSFADER"] = parse_identifier("MixerControls", "CrossfaderSlider")
                mixer_options["CROSSFADERCHANNEL"] = parse_channel("MixerControls", "CrossfaderSliderChannel")
                volume_map_mode = parse_map_mode(config_parser, "MixerControls", "VolumeMapMode")
                mixer_options["SENDMAPMODE"] = parse_map_mode(config_parser, "MixerControls", "SendsMapMode")
                mixer_options["CROSSFADERMAPMODE"] = parse_map_mode(config_parser, "MixerControls", "CrossfaderMapMode")
                if config_parser.has_option("MixerControls", "MixerButtonsToggle"):
                    if not config_parser.getboolean("MixerControls", "MixerButtonsToggle"):
                        mixer_options["NOTOGGLE"] = True
                    if config_parser.has_option("MixerControls", "InvertMuteButtonFeedback"):
                        mixer_options["INVERTMUTELEDS"] = config_parser.getboolean(
                            "MixerControls",
                            "InvertMuteButtonFeedback",
                        )
                mixer_options["NEXTBANK"] = parse_identifier("MixerControls", "NextBankButton")
                mixer_options["PREVBANK"] = parse_identifier("MixerControls", "PrevBankButton")
            for index in range(8):
                friendly_index = index + 1
                if config_parser.has_section("DeviceControls"):
                    bank_controls[f"BANK{friendly_index}"] = parse_identifier(
                        "DeviceControls",
                        f"Bank{friendly_index}Button",
                    )
                if config_parser.has_section("MixerControls"):
                    volume_controls[index] = (
                        parse_identifier("MixerControls", f"VolumeSlider{friendly_index}"),
                        parse_channel("MixerControls", f"Slider{friendly_index}Channel"),
                    )
                    trackarm_controls[index] = parse_identifier("MixerControls", f"TrackArmButton{friendly_index}")
                    mixer_options["MUTE"][index] = parse_identifier("MixerControls", f"TrackMuteButton{friendly_index}")
                    mixer_options["SOLO"][index] = parse_identifier("MixerControls", f"TrackSoloButton{friendly_index}")
                    mixer_options["SELECT"][index] = parse_identifier(
                        "MixerControls",
                        f"TrackSelectButton{friendly_index}",
                    )
                    mixer_options["SEND1"][index] = parse_identifier("MixerControls", f"Send1Knob{friendly_index}")
                    mixer_options["SEND2"][index] = parse_identifier("MixerControls", f"Send2Knob{friendly_index}")

        else:
            pass
    else:
        pass
    return type(get_name_for_script(user_path), (GenericScript,), {})(
        c_instance,
        device_map_mode,
        volume_map_mode,
        (tuple(device_controls)),
        transport_controls,
        (tuple(volume_controls)),
        (tuple(trackarm_controls)),
        bank_controls,
        controller_descriptions,
        mixer_options,
        device_component_class=DeviceComponent,
    )



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/state.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3207 bytes
from dataclasses import dataclass
from pprint import pformat
from typing import Callable, Optional
from weakref import ref

import Live


@dataclass
class StateFilters:
    key_filter = lambda k: not k.startswith("_")
    key_filter: Callable
    value_filter = lambda v: v is not None and v != ""
    value_filter: Callable


class State:
    repr_filters = StateFilters()

    def __init__(self):
        super().__init__()
        self._timers = {}

    def __repr__(self):
        return pformat((self.get_repr_data()), indent=4)

    def set_delayed(self, attr_name: str, value, delay_time: Optional[float]):
        if delay_time is None:
            setattr(self, attr_name, value)
        else:
            _self = ref(self)

            def do_setattr():
                if _self():
                    setattr(_self(), attr_name, value)
                    del _self()._timers[attr_name]

            self._timers[attr_name] = Live.Base.Timer(
                callback=do_setattr,
                interval=(int(delay_time * 1000)),
                start=True,
            )

    def get_repr_data(self):
        return State.as_dict(self, self.repr_filters)

    @staticmethod
    def as_dict(instance, state_filters=StateFilters(value_filter=(lambda _: True))):
        if isinstance(instance, dict):
            return instance
        dct = dict(vars(instance))
        keys_to_remove = set()
        for key, value in dct.items():
            if not state_filters.key_filter(key):
                keys_to_remove.add(key)
            else:
                if isinstance(value, State):
                    dct[key] = State.as_dict(value, state_filters)
            if not state_filters.value_filter(value):
                keys_to_remove.add(key)

        for key in keys_to_remove:
            del dct[key]

        return dct

    def trigger_timers(self, from_test=False):
        for timer in list(self._timers.values()):
            timer.trigger()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/loop_selector.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 8035 bytes
from ...base import depends, listens
from ...live import action, get_bar_length, liveobj_changed, liveobj_valid
from .. import Component
from ..controls import ButtonControl, control_matrix
from ..display import Renderable
from ..skin import LiveObjSkinEntry


def quantize_page_time(page_time, bar_length):
    return int(page_time / bar_length) * bar_length


class LoopSelectorComponent(Component, Renderable):
    next_page_button = ButtonControl(color="LoopSelector.Navigation", pressed_color="LoopSelector.NavigationPressed")
    prev_page_button = ButtonControl(color="LoopSelector.Navigation", pressed_color="LoopSelector.NavigationPressed")
    delete_button = ButtonControl(None)
    matrix = control_matrix(ButtonControl)

    @depends(target_track=None, sequencer_clip=None)
    def __init__(self, name="Loop_Selector", target_track=None, sequencer_clip=None, paginator=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._pressed_matrix_indices = []
        self._target_track = target_track
        self._LoopSelectorComponent__on_target_track_color_changed.subject = target_track
        self._clip = None
        self._last_bar_length = sequencer_clip.bar_length
        self._sequencer_clip = sequencer_clip
        self._LoopSelectorComponent__on_sequencer_clip_changed.subject = sequencer_clip
        self.register_slot(sequencer_clip, self._rectify_page_time, "bar_length")
        self.register_slot(sequencer_clip, self._update_matrix, "length")
        self._paginator = paginator
        self._last_page_length = paginator.page_length
        self.register_slot(paginator, self._rectify_page_time, "page_length")
        self.register_slot(paginator, self.update, "page_time")
        self.register_slot(self.song, self._update_matrix, "session_record")
        self.register_slot(self.song, self._update_matrix, "is_playing")
        self.set_clip(self._sequencer_clip.clip)

    @property
    def bar_length(self):
        return get_bar_length(clip=(self._clip))

    def set_matrix(self, matrix):
        self.matrix.set_control_element(matrix)
        self._update_matrix()

    def set_clip(self, clip):
        if liveobj_changed(clip, self._clip):
            self._LoopSelectorComponent__on_playing_position_changed.subject = clip
            self._LoopSelectorComponent__on_playing_status_changed.subject = clip
            self._clip = clip
            if clip:
                if self._paginator.can_change_page:
                    self._paginator.page_time = clip.loop_start
            self.update()

    @matrix.pressed
    def matrix(self, button):
        bar_length = self.bar_length
        if self.delete_button.is_pressed:
            if action.delete_notes_in_range(self._clip, button.index * bar_length, bar_length):
                self.notify(self.notifications.Notes.delete)
            else:
                self.notify(self.notifications.Notes.error_no_notes_to_delete)
        else:
            self._pressed_matrix_indices.append(button.index)
            if len(self._pressed_matrix_indices) > 1:
                action.set_loop_position(
                    self._clip,
                    min(self._pressed_matrix_indices) * bar_length,
                    max(self._pressed_matrix_indices) * bar_length + bar_length,
                )
            self._paginator.page_time = self._pressed_matrix_indices[0] * bar_length

    @matrix.released
    def matrix(self, button):
        if button.index in self._pressed_matrix_indices:
            self._pressed_matrix_indices.remove(button.index)

    @matrix.double_clicked
    def matrix(self, button):
        start = button.index * self.bar_length
        action.set_loop_position(self._clip, start, start + self.bar_length)

    @next_page_button.pressed
    def next_page_button(self, _):
        self._increment_page_time(1)

    @prev_page_button.pressed
    def prev_page_button(self, _):
        self._increment_page_time(-1)

    def _increment_page_time(self, delta):
        self._paginator.page_time = max(0.0, self._paginator.page_length * delta + self._paginator.page_time)

    def _rectify_page_time(self):
        if self._paginator.page_length > self._last_page_length or self.bar_length != self._last_bar_length:
            self._paginator.page_time = quantize_page_time(self._paginator.page_time, self.bar_length)
        self._last_page_length = self._paginator.page_length
        self._last_bar_length = self.bar_length
        self.update()

    def _has_clip(self):
        return liveobj_valid(self._clip)

    def update(self):
        super().update()
        self._update_matrix()
        self._update_page_buttons()

    def _update_matrix(self):
        has_clip = self._has_clip()
        for button in self.matrix:
            button.enabled = has_clip

        if has_clip:
            bar_length = self.bar_length
            loop_start = self._clip.loop_start
            loop_end = self._clip.loop_end
            playing_page = int(self._clip.playing_position / bar_length)
            editing_page = int(self._paginator.page_time / bar_length)
            for index, button in enumerate(self.matrix):
                self._update_matrix_button(
                    button,
                    index == editing_page,
                    self._clip.is_playing or self._clip.is_triggered and self.song.is_playing and index == playing_page,
                    loop_start <= button.index * bar_length < loop_end,
                )

    def _update_matrix_button(self, button, selected, playing, inside_loop):
        color = "OutsideLoop"
        if playing:
            color = "PlayheadRecord" if self.song.session_record else "Playhead"
        else:
            if inside_loop:
                color = "InsideLoopSelected" if selected else "InsideLoop"
            else:
                if selected:
                    color = "OutsideLoopSelected"
        button.color = LiveObjSkinEntry(f"LoopSelector.{color}", self._target_track.target_track)

    def _update_page_buttons(self):
        has_clip = self._has_clip()
        self.prev_page_button.enabled = has_clip and self._paginator.page_time > 0.0
        self.next_page_button.enabled = has_clip

    @listens("clip")
    def __on_sequencer_clip_changed(self):
        self.set_clip(self._sequencer_clip.clip)

    @listens("playing_position")
    def __on_playing_position_changed(self):
        self._update_matrix()

    @listens("playing_status")
    def __on_playing_status_changed(self):
        self._update_matrix()

    @listens("target_track.color")
    def __on_target_track_color_changed(self):
        self._update_matrix()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/optional.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1136 bytes
from ...base import listens
from .combo import ToggleElement


class ChoosingElement(ToggleElement):
    def __init__(self, flag=None, *a, **k):
        (super().__init__)(*a, **k)
        self._ChoosingElement__on_flag_changed.subject = flag
        self._ChoosingElement__on_flag_changed(flag.value)

    @listens("value")
    def __on_flag_changed(self, value):
        self.set_toggled(value)


class OptionalElement(ChoosingElement):
    def __init__(self, control=None, flag=None, value=None, *a, **k):
        on_control = control if value else None
        off_control = None if value else control
        (super().__init__)(a, on_control=on_control, off_control=off_control, flag=flag, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/profile.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1775 bytes
from functools import partial, wraps

ENABLE_PROFILING = False
if ENABLE_PROFILING:
    import cProfile

    PROFILER = cProfile.Profile()


def profile(fn):
    if ENABLE_PROFILING:

        @wraps(fn)
        def wrapper(self, *a, **k):
            if PROFILER:
                return PROFILER.runcall(partial(fn, self, *a, **k))
            print(f"Can not profile ({fn.__name__}), it is probably reloaded")
            return fn(*a, **k)

        return wrapper
    return fn


def dump(name="default"):
    import pstats

    fname = name + ".profile"
    PROFILER.dump_stats(fname)

    def save_human_data(sort):
        s = pstats.Stats(fname, stream=(open(f"{fname}.{sort}.txt", "w")))
        s.sort_stats(sort)
        s.print_stats()

    save_human_data("time")
    save_human_data("cumulative")



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/drum_group_scroll.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2582 bytes
from ...live import liveobj_valid
from ..display import Renderable
from .page import Pageable, PageComponent

MAX_DRUM_GROUP_SCROLL_POSITION = 28


class DrumGroupScrollComponent(PageComponent, Pageable, Renderable):
    position_count = 32
    page_length = 4
    page_offset = 1

    def __init__(self, name="Drum_Group_Scroll", *a, **k):
        self._drum_group_device = None
        (super().__init__)(a, name=name, scroll_skin_name="DrumGroup.Scroll", **k)

    def set_scroll_encoder(self, encoder):
        self._position_scroll.set_scroll_encoder(encoder)

    def set_drum_group_device(self, drum_group_device):
        self._drum_group_device = drum_group_device
        self.update()

    @property
    def position(self):
        if liveobj_valid(self._drum_group_device):
            return self._drum_group_device.view.drum_pads_scroll_position
        return 0

    @position.setter
    def position(self, index):
        if liveobj_valid(self._drum_group_device):
            self._drum_group_device.view.drum_pads_scroll_position = index

    def can_scroll_page_up(self):
        if not liveobj_valid(self._drum_group_device):
            return False
        return super().can_scroll_page_up()

    def scroll_page_up(self):
        super().scroll_page_up()
        self.notify(self.notifications.DrumGroup.Page.up)

    def scroll_page_down(self):
        super().scroll_page_down()
        self.notify(self.notifications.DrumGroup.Page.down)

    def scroll_up(self):
        super().scroll_up()
        self.notify(self.notifications.DrumGroup.Scroll.up)

    def scroll_down(self):
        super().scroll_down()
        self.notify(self.notifications.DrumGroup.Scroll.down)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/transport.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11341 bytes
from Live.Song import RecordingQuantization

from ...base import clamp, listens, sign, task
from ...live import get_bar_length, move_current_song_time
from .. import Component
from ..controls import ButtonControl, StepEncoderControl, ToggleButtonControl
from ..display import Renderable
from ..skin import OptionalSkinEntry


class TransportComponent(Component, Renderable):
    arrangement_position_encoder = StepEncoderControl(num_steps=64)
    loop_start_encoder = StepEncoderControl(num_steps=64)
    tempo_coarse_encoder = StepEncoderControl(num_steps=64)
    tempo_fine_encoder = StepEncoderControl(num_steps=64)
    play_button = ButtonControl(color="Transport.PlayOff", on_color="Transport.PlayOn")
    play_toggle_button = ToggleButtonControl(color="Transport.PlayOff", on_color="Transport.PlayOn")
    play_pause_button = ButtonControl(color="Transport.PlayOff", on_color="Transport.PlayOn")
    stop_button = ButtonControl(
        color="Transport.StopOff",
        on_color="Transport.StopOn",
        pressed_color=(OptionalSkinEntry("Transport.StopPressed")),
    )
    automation_arm_button = ToggleButtonControl(
        color="Transport.AutomationArmOff",
        on_color="Transport.AutomationArmOn",
    )
    re_enable_automation_button = ButtonControl(color="Transport.CanReEnableAutomation")
    capture_midi_button = ButtonControl(color="Transport.CanCaptureMidi")
    loop_button = ToggleButtonControl(color="Transport.LoopOff", on_color="Transport.LoopOn")
    metronome_button = ToggleButtonControl(color="Transport.MetronomeOff", on_color="Transport.MetronomeOn")
    punch_in_button = ToggleButtonControl(color="Transport.PunchOff", on_color="Transport.PunchOn")
    punch_out_button = ToggleButtonControl(color="Transport.PunchOff", on_color="Transport.PunchOn")
    tap_tempo_button = ButtonControl(color="Transport.TapTempo", pressed_color="Transport.TapTempoPressed")
    nudge_down_button = ButtonControl(color="Transport.Nudge", pressed_color="Transport.NudgePressed")
    nudge_up_button = ButtonControl(color="Transport.Nudge", pressed_color="Transport.NudgePressed")
    seek_dict = {"color": "Transport.Seek", "pressed_color": "Transport.SeekPressed", "repeat": True, "delay_time": 0}
    rewind_button = ButtonControl(**seek_dict)
    fastforward_button = ButtonControl(**seek_dict)
    prev_cue_button = ButtonControl(color="Transport.CannotJumpToCue", on_color="Transport.CanJumpToCue")
    next_cue_button = ButtonControl(color="Transport.CannotJumpToCue", on_color="Transport.CanJumpToCue")
    set_cue_button = ButtonControl(color="Transport.SetCue", pressed_color="Transport.SetCuePressed")
    record_quantize_button = ButtonControl(color="Transport.RecordQuantizeOff", on_color="Transport.RecordQuantizeOn")

    def __init__(self, name="Transport", *a, **k):
        (super().__init__)(a, name=name, **k)
        self._end_undo_step_task = self._tasks.add(task.sequence(task.wait(1.5), task.run(self.song.end_undo_step)))
        self._end_undo_step_task.kill()
        song = self.song
        self.loop_start_encoder.connect_property(
            song,
            "loop_start",
            transform=(lambda x: max(0.0, song.loop_start + get_bar_length() * sign(x))),
        )
        self.tempo_coarse_encoder.connect_property(
            song,
            "tempo",
            transform=(lambda x: clamp(song.tempo + sign(x), 20, 999)),
        )
        self.tempo_fine_encoder.connect_property(
            song,
            "tempo",
            transform=(lambda x: clamp(song.tempo + sign(x) / 10, 20, 999)),
        )
        self.play_toggle_button.connect_property(song, "is_playing")
        self.automation_arm_button.connect_property(song, "session_automation_record")
        self.loop_button.connect_property(song, "loop")
        self.metronome_button.connect_property(song, "metronome")
        self.punch_in_button.connect_property(song, "punch_in")
        self.punch_out_button.connect_property(song, "punch_out")
        self._TransportComponent__on_is_playing_changed.subject = song
        self._TransportComponent__on_is_playing_changed()
        self._TransportComponent__on_re_enable_automation_enabled_changed.subject = song
        self._TransportComponent__on_re_enable_automation_enabled_changed()
        self._TransportComponent__on_can_capture_midi_changed.subject = song
        self._TransportComponent__on_can_capture_midi_changed()
        self._TransportComponent__on_can_jump_to_prev_cue_changed.subject = song
        self._TransportComponent__on_can_jump_to_prev_cue_changed()
        self._TransportComponent__on_can_jump_to_next_cue_changed.subject = song
        self._TransportComponent__on_can_jump_to_next_cue_changed()
        self._last_record_quantization_value = RecordingQuantization.rec_q_sixtenth
        self._TransportComponent__on_record_quantization_changed.subject = song
        self._TransportComponent__on_record_quantization_changed()

    @arrangement_position_encoder.value
    def arrangement_position_encoder(self, value, _):
        move_current_song_time(self.song, sign(value))

    @metronome_button.pressed
    def metronome_button(self, button):
        self.notify(self.notifications.Transport.metronome, not button.is_on)

    @loop_button.pressed
    def loop_button(self, button):
        self.notify(self.notifications.Transport.loop, not button.is_on)

    @play_button.pressed
    def play_button(self, _):
        self.song.start_playing()

    @play_pause_button.pressed
    def play_pause_button(self, _):
        if self.song.is_playing:
            self.song.stop_playing()
        else:
            self.song.continue_playing()

    @stop_button.pressed
    def stop_button(self, _):
        self.song.stop_playing()

    @re_enable_automation_button.pressed
    def re_enable_automation_button(self, _):
        if self.song.re_enable_automation_enabled:
            self.song.re_enable_automation()

    @capture_midi_button.pressed
    def capture_midi_button(self, _):
        try:
            initial_tempo = self.song.tempo
            if self.song.can_capture_midi:
                self.song.capture_midi()
                tempo_set_by_capture = self.song.tempo != initial_tempo
                self.notify(self.notifications.Transport.midi_capture, tempo_set_by_capture, self.song.tempo)
        except RuntimeError:
            pass

    @tap_tempo_button.pressed
    def tap_tempo_button(self, _):
        self._trigger_tap_tempo()

    def _trigger_tap_tempo(self):
        if not self._end_undo_step_task.is_running:
            self.song.begin_undo_step()
        self._end_undo_step_task.restart()
        self.song.tap_tempo()

    @tap_tempo_button.released
    def tap_tempo_button(self, _):
        self.notify(self.notifications.Transport.tap_tempo, self.song.tempo)

    @nudge_down_button.value
    def nudge_down_button(self, value, _):
        self.song.nudge_down = bool(value)

    @nudge_up_button.value
    def nudge_up_button(self, value, _):
        self.song.nudge_up = bool(value)

    @rewind_button.pressed
    def rewind_button(self, _):
        move_current_song_time(self.song, -1)

    @fastforward_button.pressed
    def fastforward_button(self, _):
        move_current_song_time(self.song, 1)

    @prev_cue_button.pressed
    def prev_cue_button(self, _):
        if self.song.can_jump_to_prev_cue:
            self.song.jump_to_prev_cue()

    @next_cue_button.pressed
    def next_cue_button(self, _):
        if self.song.can_jump_to_next_cue:
            self.song.jump_to_next_cue()

    @set_cue_button.pressed
    def set_cue_button(self, _):
        self.song.set_or_delete_cue()

    @record_quantize_button.pressed
    def record_quantize_button(self, _):
        self._toggle_record_quantize()

    def _toggle_record_quantize(self):
        self.notify(self.notifications.Transport.record_quantize, not self.record_quantize_button.is_on)
        if self.song.midi_recording_quantization == RecordingQuantization.rec_q_no_q:
            self.song.midi_recording_quantization = self._last_record_quantization_value
        else:
            self.song.midi_recording_quantization = RecordingQuantization.rec_q_no_q

    @listens("is_playing")
    def __on_is_playing_changed(self):
        is_playing = self.song.is_playing
        self.play_button.is_on = is_playing
        self.play_pause_button.is_on = is_playing
        self.stop_button.is_on = is_playing

    @listens("re_enable_automation_enabled")
    def __on_re_enable_automation_enabled_changed(self):
        self.re_enable_automation_button.enabled = self.song.re_enable_automation_enabled

    @listens("can_capture_midi")
    def __on_can_capture_midi_changed(self):
        self.capture_midi_button.enabled = self.song.can_capture_midi

    @listens("can_jump_to_prev_cue")
    def __on_can_jump_to_prev_cue_changed(self):
        self.prev_cue_button.is_on = self.song.can_jump_to_prev_cue

    @listens("can_jump_to_next_cue")
    def __on_can_jump_to_next_cue_changed(self):
        self.next_cue_button.is_on = self.song.can_jump_to_next_cue

    @listens("midi_recording_quantization")
    def __on_record_quantization_changed(self):
        quantization_value = self.song.midi_recording_quantization
        quantization_on = quantization_value != RecordingQuantization.rec_q_no_q
        if quantization_on:
            self._last_record_quantization_value = quantization_value
        self.record_quantize_button.is_on = quantization_on



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/sliced_simpler.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9600 bytes
from Live.Sample import SlicingStyle

from ...base import depends, listens, task
from ...live import action, liveobj_changed, liveobj_valid
from ..controls import ButtonControl
from ..display import Renderable
from ..skin import LiveObjSkinEntry
from . import Pageable, PageComponent, PitchProvider, PlayableComponent

DEFAULT_SIMPLER_TRANSLATION_CHANNEL = 14
BASE_SLICING_NOTE = 36


class SlicedSimplerComponent(PlayableComponent, PageComponent, Pageable, PitchProvider, Renderable):
    delete_button = ButtonControl(color=None)
    position_count = 16
    page_length = 4
    page_offset = 0

    @depends(target_track=None)
    def __init__(
        self,
        name="Sliced_Simpler",
        translation_channel=DEFAULT_SIMPLER_TRANSLATION_CHANNEL,
        target_track=None,
        *a,
        **k,
    ):
        self._position = 0
        self._simpler_device = None
        (super().__init__)(a, name=name, scroll_skin_name="SlicedSimpler.Scroll", **k)
        self._target_track = target_track
        self._translation_channel = translation_channel

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, index):
        self._position = index
        self.notify_position()
        self._update_led_feedback()
        self._update_note_translations()

    def can_scroll_page_up(self):
        if not liveobj_valid(self._simpler_device) or (self._position + 1) * self.page_length > len(self._slices()):
            return False
        return super().can_scroll_page_up()

    def set_scroll_encoder(self, encoder):
        self._position_scroll.set_scroll_encoder(encoder)

    def set_simpler_device(self, simpler_device):
        if liveobj_changed(simpler_device, self._simpler_device):
            self._simpler_device = simpler_device
            self._SlicedSimplerComponent__on_file_changed.subject = simpler_device
            self._SlicedSimplerComponent__on_selected_slice_changed.subject = simpler_device
            self._SlicedSimplerComponent__on_pad_slicing_changed.subject = simpler_device
            sample = simpler_device.sample if liveobj_valid(simpler_device) else None
            self._SlicedSimplerComponent__on_slices_changed.subject = sample
            self._SlicedSimplerComponent__on_slicing_style_changed.subject = sample
            self._SlicedSimplerComponent__on_track_color_changed.subject = (
                self._target_track.target_track if simpler_device else None
            )
            self.update()
        if not liveobj_valid(self._simpler_device):
            self.update()

    def scroll_page_up(self):
        super().scroll_page_up()
        self.notify(self.notifications.Simpler.Page.up)

    def scroll_page_down(self):
        super().scroll_page_down()
        self.notify(self.notifications.Simpler.Page.down)

    def scroll_up(self):
        super().scroll_up()
        self.notify(self.notifications.Simpler.Scroll.up)

    def scroll_down(self):
        super().scroll_down()
        self.notify(self.notifications.Simpler.Scroll.down)

    def _on_matrix_pressed(self, button):
        if self._simpler_setup_is_valid():
            slice_index = self._coordinate_to_slice_index(button.coordinate)
            if self.delete_button.is_pressed:
                button.color = "SlicedSimpler.PadAction"
                if action.delete_notes_with_pitch(self._target_track.target_clip, button.identifier):
                    self.notify(self.notifications.Simpler.Slice.delete_notes, slice_index + 1)
                else:
                    self._delete_slice_at_index(slice_index)
            if self.select_button.is_pressed:
                self._select_slice_at_index(slice_index)
                super()._on_matrix_pressed(button)

    @delete_button.value
    def delete_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    def _any_modifier_pressed(self):
        return self.select_button.is_pressed or self.delete_button.is_pressed

    def _set_control_pads_from_script(self, takeover_pads):
        super()._set_control_pads_from_script(takeover_pads or self._any_modifier_pressed())

    def _select_slice_at_index(self, index):
        slices = self._slices()
        if len(slices) > index:
            try:
                if liveobj_changed(self._simpler_device.view.selected_slice, slices[index]):
                    self._simpler_device.view.selected_slice = slices[index]
                    self.notify(self.notifications.Simpler.Slice.select, index + 1)
            except RuntimeError:
                pass

    def _delete_slice_at_index(self, index):
        slices = self._slices()
        if len(slices) > index:
            self._simpler_device.sample.remove_slice(slices[index])
            self.notify(self.notifications.Simpler.Slice.delete, index + 1)

    def _slices(self):
        if self._simpler_setup_is_valid():
            return self._simpler_device.sample.slices
        return []

    def _selected_slice(self):
        if self._simpler_setup_is_valid():
            return self._simpler_device.view.selected_slice
        return -1

    def _should_show_next_slice(self, index, length_of_slices):
        return (
            self._simpler_setup_is_valid()
            and index == length_of_slices
            and self._simpler_device.pad_slicing
            and self._simpler_device.sample.slicing_style == SlicingStyle.manual
        )

    def _simpler_setup_is_valid(self):
        return liveobj_valid(self._simpler_device) and liveobj_valid(self._simpler_device.sample)

    def update(self):
        if self._position * self.page_length > len(self._slices()):
            self.position = 0
        super().update()
        self._update_provided_pitches()
        self._update_led_feedback()

    def _update_provided_pitches(self):
        slices = self._slices()
        selected_slice = self._selected_slice()
        if slices:
            offset = slices.index(selected_slice) if selected_slice in slices else 0
            self.pitches = [BASE_SLICING_NOTE + offset]

    def _update_button_color(self, button):
        index = self._coordinate_to_slice_index(button.coordinate)
        slices = self._slices()
        length_of_slices = len(slices)
        color = "SlicedSimpler.NoSlice"
        if index < length_of_slices:
            color = (
                "SlicedSimpler.SliceSelected"
                if slices[index] == self._selected_slice()
                else "SlicedSimpler.SliceNotSelected"
            )
        else:
            if self._should_show_next_slice(index, length_of_slices):
                color = "SlicedSimpler.NextSlice"
        button.color = LiveObjSkinEntry(color, self._target_track.target_track)

    def _note_translation_for_button(self, button):
        identifier = BASE_SLICING_NOTE + self._coordinate_to_slice_index(button.coordinate)
        return (identifier, self._translation_channel)

    def _coordinate_to_slice_index(self, coordinate):
        y, x = coordinate
        y = self.height - y - 1
        y += self._position
        y += self.height if x >= 4 else 0
        return x % 4 + y * 4

    @listens("sample")
    def __on_file_changed(self):
        self._SlicedSimplerComponent__on_slices_changed.subject = (
            self._simpler_device.sample if liveobj_valid(self._simpler_device) else None
        )
        self.update()

    @listens("view.selected_slice")
    def __on_selected_slice_changed(self):
        self.update()

    @listens("pad_slicing")
    def __on_pad_slicing_changed(self):
        self.update()

    @listens("slices")
    def __on_slices_changed(self):
        self.update()

    @listens("slicing_style")
    def __on_slicing_style_changed(self):
        def set_pad_slicing():
            self._simpler_device.pad_slicing = self._simpler_device.sample.slicing_style == SlicingStyle.manual

        self._tasks.add(task.sequence(task.delay(1), task.run(set_pad_slicing)))

    @listens("color_index")
    def __on_track_color_changed(self):
        self._update_led_feedback()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/clip_slot.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 10057 bytes
import Live

from ...base import listens, liveobj_valid
from ..component import Component
from ..control import ButtonControl


def find_nearest_color(rgb_table, src_hex_color):
    def hex_to_channels(color_in_hex):
        return ((color_in_hex & 16711680) >> 16, (color_in_hex & 65280) >> 8, color_in_hex & 255)

    def squared_distance(color):
        return sum([(a - b) ** 2 for a, b in zip(hex_to_channels(src_hex_color), hex_to_channels(color[1]))])

    return min(rgb_table, key=squared_distance)[0]


def is_button_pressed(button):
    if button:
        return button.is_pressed()
    return False


class ClipSlotComponent(Component):
    launch_button = ButtonControl()

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._clip_slot = None
        self._triggered_to_play_color = "Session.ClipTriggeredPlay"
        self._triggered_to_record_color = "Session.ClipTriggeredRecord"
        self._started_value = "Session.ClipStarted"
        self._recording_color = "Session.ClipRecording"
        self._stopped_value = "Session.ClipStopped"
        self._clip_palette = []
        self._clip_rgb_table = None
        self._record_button_color = "Session.RecordButton"
        self._empty_slot_color = "Session.ClipEmpty"
        self._delete_button = None
        self._select_button = None
        self._duplicate_button = None

    def set_clip_slot(self, clip_slot):
        self._clip_slot = clip_slot
        self._update_clip_property_slots()
        self._ClipSlotComponent__on_slot_triggered_changed.subject = clip_slot
        self._ClipSlotComponent__on_slot_playing_state_changed.subject = clip_slot
        self._ClipSlotComponent__on_clip_state_changed.subject = clip_slot
        self._ClipSlotComponent__on_controls_other_clips_changed.subject = clip_slot
        self._ClipSlotComponent__on_has_stop_button_changed.subject = clip_slot
        self._ClipSlotComponent__on_clip_slot_color_changed.subject = clip_slot
        track = clip_slot.canonical_parent if clip_slot else None
        if track:
            if track in self.song.tracks:
                self._ClipSlotComponent__on_arm_value_changed.subject = track
                self._ClipSlotComponent__on_implicit_arm_value_changed.subject = track
                self._ClipSlotComponent__on_input_routing_type_changed.subject = track
        self.update()

    def set_launch_button(self, button):
        self.launch_button.set_control_element(button)
        self.update()

    def set_delete_button(self, button):
        self._delete_button = button

    def set_select_button(self, button):
        self._select_button = button

    def set_duplicate_button(self, button):
        self._duplicate_button = button

    def set_clip_palette(self, palette):
        self._clip_palette = palette

    def set_clip_rgb_table(self, rgb_table):
        self._clip_rgb_table = rgb_table

    def has_clip(self):
        return self._clip_slot.has_clip

    def update(self):
        super().update()
        self._update_launch_button_color()

    def _update_launch_button_color(self):
        if self.is_enabled():
            value_to_send = self._empty_slot_color
            if liveobj_valid(self._clip_slot):
                track = self._clip_slot.canonical_parent
                slot_or_clip = self._clip_slot.clip if self.has_clip() else self._clip_slot
                value_to_send = self._feedback_value(track, slot_or_clip)
            self.launch_button.color = value_to_send

    def _color_value(self, slot_or_clip):
        color = slot_or_clip.color
        try:
            return self._clip_palette[color]
        except (KeyError, IndexError):
            if self._clip_rgb_table is not None:
                return find_nearest_color(self._clip_rgb_table, color)
            else:
                return self._stopped_value

    def _track_is_armed(self, track):
        return liveobj_valid(track) and track.can_be_armed and any([track.arm, track.implicit_arm])

    def _feedback_value(self, track, slot_or_clip):
        if slot_or_clip.is_triggered:
            if slot_or_clip.will_record_on_start:
                return self._triggered_to_record_color
            return self._triggered_to_play_color
        if slot_or_clip.is_playing:
            if slot_or_clip.is_recording:
                return self._recording_color
            return self._started_value
        if slot_or_clip.color is not None:
            return self._color_value(slot_or_clip)
        if getattr(slot_or_clip, "controls_other_clips", True):
            return self._stopped_value
        if self._track_is_armed(track):
            if self._clip_slot.has_stop_button:
                if self._record_button_color is not None:
                    return self._record_button_color
        return self._empty_slot_color

    def _update_clip_property_slots(self):
        clip = self._clip_slot.clip if self._clip_slot else None
        self._ClipSlotComponent__on_clip_playing_state_changed.subject = clip
        self._ClipSlotComponent__on_recording_state_changed.subject = clip
        self._ClipSlotComponent__on_clip_color_changed.subject = clip

    @listens("has_clip")
    def __on_clip_state_changed(self):
        self._update_clip_property_slots()
        self._update_launch_button_color()

    @listens("controls_other_clips")
    def __on_controls_other_clips_changed(self):
        self._update_clip_property_slots()
        self._update_launch_button_color()

    @listens("color")
    def __on_clip_color_changed(self):
        self._update_launch_button_color()

    @listens("color")
    def __on_clip_slot_color_changed(self):
        self._update_launch_button_color()

    @listens("playing_status")
    def __on_slot_playing_state_changed(self):
        self._update_launch_button_color()

    @listens("playing_status")
    def __on_clip_playing_state_changed(self):
        self._update_launch_button_color()

    @listens("is_recording")
    def __on_recording_state_changed(self):
        self._update_launch_button_color()

    @listens("arm")
    def __on_arm_value_changed(self):
        self._update_launch_button_color()

    @listens("implicit_arm")
    def __on_implicit_arm_value_changed(self):
        self._update_launch_button_color()

    @listens("input_routing_type")
    def __on_input_routing_type_changed(self):
        self._update_launch_button_color()

    @listens("has_stop_button")
    def __on_has_stop_button_changed(self):
        self._update_launch_button_color()

    @listens("is_triggered")
    def __on_slot_triggered_changed(self):
        if not self.has_clip():
            self._update_launch_button_color()

    @launch_button.pressed
    def launch_button(self, button):
        self._on_launch_button_pressed()

    def _on_launch_button_pressed(self):
        if is_button_pressed(self._select_button):
            self._do_select_clip(self._clip_slot)
        else:
            if liveobj_valid(self._clip_slot):
                if is_button_pressed(self._duplicate_button):
                    self._do_duplicate_clip()
                else:
                    if is_button_pressed(self._delete_button):
                        self._do_delete_clip()
                    else:
                        self._do_launch_clip(True)
                        self._show_launched_clip_as_highlighted_clip()

    @launch_button.released
    def launch_button(self, button):
        self._on_launch_button_released()

    def _on_launch_button_released(self):
        if self.launch_button.is_momentary:
            if not is_button_pressed(self._select_button) or liveobj_valid(self._clip_slot):
                if not is_button_pressed(self._duplicate_button):
                    if not is_button_pressed(self._delete_button):
                        self._do_launch_clip(False)

    def _do_delete_clip(self):
        if self._clip_slot:
            if self._clip_slot.has_clip:
                self._clip_slot.delete_clip()
                self._on_clip_deleted()

    def _on_clip_deleted(self):
        pass

    def _do_select_clip(self, clip_slot):
        if liveobj_valid(self._clip_slot):
            if self.song.view.highlighted_clip_slot != self._clip_slot:
                self.song.view.highlighted_clip_slot = self._clip_slot
                self._on_slot_selected()

    def _on_slot_selected(self):
        pass

    def _do_duplicate_clip(self):
        if not self._clip_slot or self._clip_slot.has_clip:
            try:
                track = self._clip_slot.canonical_parent
                track.duplicate_clip_slot(list(track.clip_slots).index(self._clip_slot))
                self._on_clip_duplicated()
            except Live.Base.LimitationError:
                pass
            except RuntimeError:
                pass

    def _on_clip_duplicated(self):
        pass

    def _do_launch_clip(self, fire_state):
        object_to_launch = self._clip_slot
        if self.has_clip():
            object_to_launch = self._clip_slot.clip
        object_to_launch.set_fire_button_state(fire_state)

    def _show_launched_clip_as_highlighted_clip(self):
        song = self.song
        if song.select_on_launch:
            if self._clip_slot != song.view.highlighted_clip_slot:
                self.song.view.highlighted_clip_slot = self._clip_slot



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_UserScript/DeviceComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4915 bytes
from _Framework.DeviceComponent import DeviceComponent as DeviceComponentBase
from ableton.v2.base import liveobj_valid, nop


class DeviceComponent(DeviceComponentBase):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._num_banks_to_control = 1
        self._empty_control_slots = self.register_slot_manager()

    def set_parameter_controls(self, controls):
        self._num_banks_to_control = 2 if controls and (controls.width() > 8) else 1
        super().set_parameter_controls(controls)

    def _bank_up_value(self, value):
        if self.is_enabled():
            if value:
                if self._bank_down_button is None:
                    self._cycle_bank_index()
                else:
                    if self._can_bank_up():
                        self._increment_bank_index(1)

    def _bank_down_value(self, value):
        if self.is_enabled():
            if value:
                if self._can_bank_down():
                    self._increment_bank_index(-1)

    def _increment_bank_index(self, delta):
        if liveobj_valid(self._device):
            delta = self._num_banks_to_control * delta
            self._bank_name = ""
            if self._bank_index is None:
                self._bank_index = 0
            else:
                self._bank_index = max(0, min(self._number_of_parameter_banks() - 1, self._bank_index + delta))
            self.update()

    def _cycle_bank_index(self):
        if liveobj_valid(self._device):
            self._bank_name = ""
            if self._bank_index is None:
                self._bank_index = 0
            else:
                self._bank_index = (self._bank_index + 1) % self._number_of_parameter_banks()
            self.update()

    def _can_bank_up(self):
        return (
            self._bank_index is None
            or self._number_of_parameter_banks() > self._bank_index + self._num_banks_to_control
        )

    def _can_bank_down(self):
        return self._bank_index is None or self._bank_index > 0

    def update(self):
        self._empty_control_slots.disconnect()
        super().update()

    def _assign_parameters(self):
        self._bank_name, bank = self._current_bank_details()
        for control, parameter in zip(self._parameter_controls, bank):
            if control:
                if liveobj_valid(parameter):
                    control.connect_to(parameter)
                else:
                    control.release_parameter()
                    self._empty_control_slots.register_slot(control, nop, "value")

        self._release_parameters(self._parameter_controls[len(bank) :])

    def _release_parameters(self, controls):
        for control in controls or []:
            if control:
                control.release_parameter()
                self._empty_control_slots.register_slot(control, nop, "value")

    def _current_bank_details(self):
        bank_name = self._bank_name
        bank = []
        best_of = self._best_of_parameter_bank()
        banks = self._parameter_banks()
        if banks:
            multiple_banks = self._num_banks_to_control > 1
            if self._bank_index is not None:
                if self._is_banking_enabled() or multiple_banks or not best_of:
                    index = self._bank_index if self._bank_index is not None else 0
                    bank = list(banks[index])
                    bank_name = self._parameter_bank_names()[index]
                    if not multiple_banks or index < self._number_of_parameter_banks() - 1:
                        bank.extend(banks[index + 1])
                        bank_name += f"  /  {self._parameter_bank_names()[index + 1]}"
                else:
                    bank = best_of
                    bank_name = "Best of Parameters"
        return (bank_name, bank)

    def _update_device_bank_nav_buttons(self):
        if self.is_enabled():
            if self._bank_up_button:
                if liveobj_valid(self._device):
                    if self._bank_down_button is None:
                        self._bank_up_button.set_light(self._number_of_parameter_banks() > 1)
                    else:
                        self._bank_up_button.set_light(self._can_bank_up())
                else:
                    self._bank_up_button.set_light(False)
            if self._bank_down_button:
                self._bank_down_button.set_light(liveobj_valid(self._device) and self._can_bank_down())



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/parameter_provider.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1758 bytes
from ableton.v2.base import EventObject, NamedTuple, liveobj_valid

DISCRETE_PARAMETERS_DICT = {"GlueCompressor": ("Ratio", "Attack", "Release", "Peak Clip In")}


def is_parameter_quantized(parameter, parent_device):
    is_quantized = False
    if liveobj_valid(parameter):
        device_class = getattr(parent_device, "class_name", None)
        is_quantized = (parameter.is_quantized) or (
            (device_class in DISCRETE_PARAMETERS_DICT) and (parameter.name in DISCRETE_PARAMETERS_DICT[device_class])
        )
    return is_quantized


class ParameterInfo(NamedTuple):
    parameter = None
    default_encoder_sensitivity = None
    fine_grain_encoder_sensitivity = None

    def __init__(self, name=None, *a, **k):
        (super().__init__)(a, _overridden_name=name, **k)

    @property
    def name(self):
        actual_name = self.parameter.name if liveobj_valid(self.parameter) else ""
        return self._overridden_name or actual_name

    def __eq__(self, other_info):
        if not isinstance(other_info, ParameterInfo):
            return NotImplemented
        return super().__eq__(other_info) and self.name == other_info.name

    def __hash__(self):
        return hash(
            (
                self._overridden_name,
                self.parameter,
                self.default_encoder_sensitivity,
                self.fine_grain_encoder_sensitivity,
            ),
        )


class ParameterProvider(EventObject):
    __events__ = ("parameters",)

    @property
    def parameters(self):
        return []



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/session.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11924 bytes
from itertools import count, zip_longest

from ...base import const, depends, inject, listenable_property, listens, listens_group
from ...live import liveobj_valid
from .. import Component
from ..controls import ButtonControl, control_list
from ..display import Renderable
from . import ClipboardComponent, SceneComponent


class SessionComponent(Component, Renderable):
    stop_all_clips_button = ButtonControl(color="Session.StopAllClips", pressed_color="Session.StopAllClipsPressed")
    stop_track_clip_buttons = control_list(ButtonControl)
    _session_component_ends_initialisation = True

    @depends(session_ring=None)
    def __init__(
        self,
        name="Session",
        session_ring=None,
        scene_component_type=None,
        clip_slot_component_type=None,
        clipboard_component_type=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._session_ring = session_ring
        scene_component_type = scene_component_type or SceneComponent
        create_scene = lambda: scene_component_type(
            parent=self,
            session_ring=(self._session_ring),
            clip_slot_component_type=clip_slot_component_type,
        )
        clipboard_component_type = clipboard_component_type or ClipSlotClipboardComponent
        self._clipboard = clipboard_component_type(parent=self)
        with inject(clipboard=(const(self._clipboard))).everywhere():
            self._selected_scene = create_scene()
            self._scenes = [create_scene() for _ in range(self._session_ring.num_scenes)]
        self.register_slot(self._session_ring, self._reassign_scenes, "scenes")
        self.register_slot(self._session_ring, self._reassign_tracks, "tracks")
        self._SessionComponent__on_offsets_changed.subject = self._session_ring
        self._SessionComponent__on_selected_scene_changed.subject = self.song.view
        self.stop_track_clip_buttons.control_count = session_ring.num_tracks
        if self._session_component_ends_initialisation:
            self._end_initialisation()

    @listenable_property
    def clipboard(self):
        return self._clipboard

    def scene(self, index):
        return self._scenes[index]

    def selected_scene(self):
        return self._selected_scene

    def set_clip_launch_buttons(self, buttons):
        for y, scene in enumerate(self._scenes):
            for x in range(self._session_ring.num_tracks):
                button = buttons.get_button(y, x) if buttons else None
                scene.clip_slot(x).set_launch_button(button)

    def set_scene_launch_buttons(self, buttons):
        for scene, button in zip_longest(self._scenes, buttons or []):
            scene.set_launch_button(button)

    def set_stop_track_clip_buttons(self, buttons):
        self.stop_track_clip_buttons.set_control_element(buttons)
        self._update_stop_track_clip_buttons()

    def set_clip_slot_select_button(self, button):
        self.set_modifier_button(button, "select_button", clip_slots_only=True)

    def set_select_button(self, button):
        self.set_modifier_button(button, "select_button")

    def set_delete_button(self, button):
        self.set_modifier_button(button, "delete_button")

    def set_duplicate_button(self, button):
        self.set_modifier_button(button, "duplicate_button")

    def set_copy_button(self, button):
        self._clipboard.set_copy_button(button)
        self.set_modifier_button(button, "copy_button", clip_slots_only=True)

    def set_modifier_button(self, button, name, clip_slots_only=False):
        for y in range(self._session_ring.num_scenes):
            scene = self.scene(y)
            if not clip_slots_only:
                getattr(scene, name).set_control_element(button)
            else:
                for x in range(self._session_ring.num_tracks):
                    getattr(scene.clip_slot(x), name).set_control_element(button)

    def __getattr__(self, name):
        if name.startswith("set_selected_scene_launch_button"):
            return self.selected_scene().set_launch_button
        if name.startswith("set_scene_"):
            if name.endswith("_launch_button"):
                return self.scene(int(name.split("_")[-3])).set_launch_button
        raise AttributeError

    @stop_all_clips_button.pressed
    def stop_all_clips_button(self, _):
        self.song.stop_all_clips()

    @stop_track_clip_buttons.pressed
    def stop_track_clip_buttons(self, button):
        self._session_ring.tracks[button.index].stop_all_clips()

    def _end_initialisation(self):
        self._SessionComponent__on_selected_scene_changed()
        self._reassign_tracks_and_scenes()

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_stop_track_clip_buttons()
            self._reassign_tracks_and_scenes()

    def _reassign_tracks_and_scenes(self):
        self._reassign_tracks()
        self._reassign_scenes()

    def _reassign_scenes(self):
        scenes = self.song.scenes
        for index, scene in enumerate(self._scenes):
            scene_index = self._session_ring.scene_offset + index
            scene.set_scene(scenes[scene_index] if len(scenes) > scene_index else None)

    def _reassign_tracks(self):
        tracks = self._session_ring.tracks
        self._SessionComponent__on_fired_slot_index_changed.replace_subjects(tracks, count())
        self._SessionComponent__on_playing_slot_index_changed.replace_subjects(tracks, count())
        self._update_stop_track_clip_buttons()

    def _update_stop_track_clip_buttons(self):
        if self.is_enabled():
            for index in range(self._session_ring.num_tracks):
                self._update_stop_clips_led(index)

    def _update_stop_clips_led(self, index):
        if index < self.stop_track_clip_buttons.control_count:
            tracks = self._session_ring.tracks
            track = tracks[index] if index < len(tracks) else None
            button = self.stop_track_clip_buttons[index]
            if liveobj_valid(track) and track.clip_slots:
                button.enabled = True
                if track.fired_slot_index == -2:
                    button.color = "Session.StopClipTriggered"
                else:
                    if track.playing_slot_index >= 0:
                        button.color = "Session.StopClip"
                    else:
                        button.color = "Session.StopClipDisabled"
            else:
                button.enabled = False

    @listens("offset")
    def __on_offsets_changed(self, *_):
        if self.is_enabled():
            self._reassign_tracks_and_scenes()

    @listens("selected_scene")
    def __on_selected_scene_changed(self):
        if self._selected_scene is not None:
            self._selected_scene.set_scene(self.song.view.selected_scene)

    @listens_group("fired_slot_index")
    def __on_fired_slot_index_changed(self, track_index):
        self._update_stop_clips_led(track_index)

    @listens_group("playing_slot_index")
    def __on_playing_slot_index_changed(self, track_index):
        self._update_stop_clips_led(track_index)


class ClipSlotClipboardComponent(ClipboardComponent):
    def _do_copy(self, obj):
        if not liveobj_valid(obj):
            return None
        if obj.is_group_slot:
            self.notify(self.notifications.Clip.CopyPaste.error_copy_from_group_slot)
            return None
        else:
            if not liveobj_valid(obj.clip):
                self.notify(self.notifications.Clip.CopyPaste.error_copy_from_empty_slot)
                return None
            else:
                if obj.clip.is_recording:
                    self.notify(self.notifications.Clip.CopyPaste.error_copy_recording_clip)
                    return None
                else:
                    self.notify(self.notifications.Clip.CopyPaste.copy, obj)
                    return obj

    def _do_paste(self, obj):
        if not liveobj_valid(obj):
            return False
        if obj.is_group_slot:
            self.notify(self.notifications.Clip.CopyPaste.error_paste_to_group_slot)
            return False
        source_is_audio = self._source_obj.clip.is_audio_clip
        destination_track = obj.canonical_parent
        if source_is_audio and not destination_track.has_audio_input:
            self.notify(self.notifications.Clip.CopyPaste.error_paste_audio_to_midi)
        else:
            if not source_is_audio or destination_track.has_audio_input:
                self.notify(self.notifications.Clip.CopyPaste.error_paste_midi_to_audio)
            else:
                self._source_obj.duplicate_clip_to(obj)
                self.notify(self.notifications.Clip.CopyPaste.paste, self._source_obj)
                return True
        return False

    def _is_source_valid(self):
        return liveobj_valid(self._source_obj) and self._source_obj.has_clip



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4150 bytes
from ableton.v2.control_surface import (
    BANK_FORMAT,
    BANK_MAIN_KEY,
    BANK_PARAMETERS_KEY,
    MIDI_CC_TYPE,
    MIDI_NOTE_TYPE,
    MIDI_PB_TYPE,
    MIDI_SYSEX_TYPE,
    CompoundElement,
    ControlElement,
    DeviceBankRegistry,
    DeviceProvider,
    EnumWrappingParameter,
    InputControlElement,
    NotifyingControlElement,
    NotifyingList,
    ParameterProvider,
    PrioritizedResource,
    RelativeInternalParameter,
    SharedResource,
    all_parameters,
    find_instrument_devices,
    find_instrument_meeting_requirement,
    use,
)
from ableton.v2.control_surface.default_bank_definitions import BANK_DEFINITIONS as V2_BANK_DEFINITIONS
from ableton.v2.control_surface.defaults import DOUBLE_CLICK_DELAY
from ableton.v2.control_surface.elements.encoder import ABSOLUTE_MAP_MODES
from ableton.v2.control_surface.input_control_element import ScriptForwarding

from . import midi
from .banking_util import DEFAULT_BANK_SIZE, BankingInfo, create_parameter_bank
from .colors import STANDARD_COLOR_PALETTE, STANDARD_FALLBACK_COLOR_TABLE, BasicColors
from .component import Component
from .consts import (
    ACTIVE_PARAMETER_TIMEOUT,
    DEFAULT_PRIORITY,
    HIGH_PRIORITY,
    LOW_PRIORITY,
    M4L_PRIORITY,
    MOMENTARY_DELAY,
)
from .control_surface import ControlSurface, create_control_surface
from .control_surface_specification import ControlSurfaceSpecification
from .default_bank_definitions import BANK_DEFINITIONS
from .default_skin import create_skin, default_skin
from .elements_base import (
    ElementsBase,
    MapMode,
    create_button,
    create_combo_element,
    create_encoder,
    create_matrix_identifiers,
    create_sysex_element,
    create_sysex_sending_button,
)
from .identification import IdentificationComponent
from .instrument_finder import InstrumentFinderComponent
from .layer import Layer
from .parameter_info import ParameterInfo
from .parameter_mapping_sensitivities import (
    DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY,
    DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY,
    FINE_GRAIN_SENSITIVITY_FACTOR,
    parameter_mapping_sensitivities,
)
from .session_ring_selection_linking import SessionRingSelectionLinking
from .skin import LiveObjSkinEntry, OptionalSkinEntry, Skin, merge_skins

__all__ = (
    "ABSOLUTE_MAP_MODES",
    "ACTIVE_PARAMETER_TIMEOUT",
    "BANK_DEFINITIONS",
    "BANK_FORMAT",
    "BANK_MAIN_KEY",
    "BANK_PARAMETERS_KEY",
    "DEFAULT_BANK_SIZE",
    "DEFAULT_CONTINUOUS_PARAMETER_SENSITIVITY",
    "DEFAULT_QUANTIZED_PARAMETER_SENSITIVITY",
    "DEFAULT_PRIORITY",
    "DOUBLE_CLICK_DELAY",
    "FINE_GRAIN_SENSITIVITY_FACTOR",
    "HIGH_PRIORITY",
    "LOW_PRIORITY",
    "M4L_PRIORITY",
    "MIDI_CC_TYPE",
    "MIDI_NOTE_TYPE",
    "MIDI_PB_TYPE",
    "MIDI_SYSEX_TYPE",
    "MOMENTARY_DELAY",
    "STANDARD_COLOR_PALETTE",
    "STANDARD_FALLBACK_COLOR_TABLE",
    "V2_BANK_DEFINITIONS",
    "BankingInfo",
    "BasicColors",
    "Component",
    "CompoundElement",
    "ControlElement",
    "ControlSurface",
    "ControlSurfaceSpecification",
    "DeviceBankRegistry",
    "DeviceProvider",
    "ElementsBase",
    "EnumWrappingParameter",
    "IdentificationComponent",
    "InputControlElement",
    "InstrumentFinderComponent",
    "Layer",
    "LiveObjSkinEntry",
    "MapMode",
    "NotifyingControlElement",
    "NotifyingList",
    "OptionalSkinEntry",
    "ParameterInfo",
    "ParameterProvider",
    "PrioritizedResource",
    "RelativeInternalParameter",
    "ScriptForwarding",
    "SessionRingSelectionLinking",
    "SharedResource",
    "Skin",
    "all_parameters",
    "create_button",
    "create_combo_element",
    "create_control_surface",
    "create_encoder",
    "create_matrix_identifiers",
    "create_parameter_bank",
    "create_skin",
    "create_sysex_element",
    "create_sysex_sending_button",
    "default_skin",
    "find_instrument_devices",
    "find_instrument_meeting_requirement",
    "merge_skins",
    "midi",
    "parameter_mapping_sensitivities",
    "use",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/radio_button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1778 bytes
from ...base import nop
from .button import ButtonControlBase
from .control import control_color, control_event


class RadioButtonControl(ButtonControlBase):
    checked = control_event("checked")

    class State(ButtonControlBase.State):
        unchecked_color = control_color("DefaultButton.Off")
        checked_color = control_color("DefaultButton.On")

        def __init__(self, unchecked_color=None, checked_color=None, *a, **k):
            (super(RadioButtonControl.State, self).__init__)(*a, **k)
            if unchecked_color is not None:
                self.unchecked_color = unchecked_color
            if checked_color is not None:
                self.checked_color = checked_color
            self._checked = False
            self._on_checked = nop

        @property
        def is_checked(self):
            return self._checked

        @is_checked.setter
        def is_checked(self, value):
            if self._checked != value:
                self._checked = value
                if self._checked:
                    self._on_checked()
                self._send_current_color()

        def _send_button_color(self):
            self._control_element.set_light(self.checked_color if self._checked else self.unchecked_color)

        def _on_pressed(self):
            super(RadioButtonControl.State, self)._on_pressed()
            if not self._checked:
                self._checked = True
                self._notify_checked()

        def _notify_checked(self):
            if self._checked:
                self._call_listener("checked")
                self._on_checked()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/identification.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9099 bytes
import logging
from abc import ABC, abstractmethod

from ..base import depends, listenable_property, nop, task
from . import MIDI_CC_TYPE, MIDI_NOTE_TYPE, MIDI_PB_TYPE, Component, InputControlElement, midi
from .controls import InputControl
from .display import Renderable
from .elements import SysexElement

logger = logging.getLogger(__name__)
NON_REALTIME_HEADER = (midi.SYSEX_START, midi.SYSEX_NON_REALTIME)
STANDARD_RESPONSE_BYTES = (midi.SYSEX_GENERAL_INFO, midi.SYSEX_IDENTITY_RESPONSE_ID)
PRODUCT_ID_BYTES_START_INDEX = 3


def status_byte_to_msg_type(status_byte):
    status_byte = 240 & status_byte
    msg_type = MIDI_CC_TYPE
    if status_byte == midi.NOTE_ON_STATUS:
        msg_type = MIDI_NOTE_TYPE
    else:
        if status_byte == midi.PB_STATUS:
            msg_type = MIDI_PB_TYPE
    return msg_type


def create_responder(identity_response_id_bytes, custom_identity_response):
    if identity_response_id_bytes is not None:
        return StandardResponder(identity_response_id_bytes)
    if custom_identity_response[0] == midi.SYSEX_START:
        return CustomSysexResponder(custom_identity_response)
    return PlainMidiResponder(custom_identity_response)


class Responder(ABC):
    def __init__(self, response_bytes, *a, **k):
        (super().__init__)(*a, **k)
        self._response_bytes = tuple(response_bytes)

    @abstractmethod
    def create_response_element(self):
        pass

    @abstractmethod
    def is_valid_response(self, response_bytes):
        pass


class CustomSysexResponder(Responder):
    def create_response_element(self):
        return SysexElement(sysex_identifier=(self._response_bytes))

    def is_valid_response(self, _):
        return True


class PlainMidiResponder(Responder):
    def __init__(self, response_bytes, *a, **k):
        (super().__init__)(response_bytes, *a, **k)
        self._expected_response_value_bytes = midi.extract_value(response_bytes) if len(response_bytes) == 3 else None

    def create_response_element(self):
        first_byte = self._response_bytes[0]
        element = InputControlElement(
            msg_type=(status_byte_to_msg_type(first_byte)),
            channel=(15 & first_byte),
            identifier=(self._response_bytes[1]),
        )
        element.reset = nop
        return element

    def is_valid_response(self, response_bytes):
        return self._expected_response_value_bytes is None or self._expected_response_value_bytes == response_bytes


class StandardResponder(Responder):
    def create_response_element(self):
        return SysexElement(sysex_identifier=NON_REALTIME_HEADER)

    def is_valid_response(self, response_bytes):
        if response_bytes[1:PRODUCT_ID_BYTES_START_INDEX] == STANDARD_RESPONSE_BYTES:
            product_id_bytes = self._extract_product_id_bytes(response_bytes)
            if product_id_bytes != self._response_bytes:
                raise IdentityResponseError(expected_bytes=(self._response_bytes), actual_bytes=product_id_bytes)
            return True
        return False

    def _extract_product_id_bytes(self, midi_bytes):
        return midi_bytes[PRODUCT_ID_BYTES_START_INDEX : PRODUCT_ID_BYTES_START_INDEX + len(self._response_bytes)]


class IdentityResponseError(Exception):
    def __init__(self, expected_bytes=None, actual_bytes=None):
        super().__init__(
            f"Hardware controller responded with wrong identity bytes: ({expected_bytes} != {actual_bytes}).",
        )


class IdentificationComponent(Component, Renderable):
    identity_response_control = InputControl()
    is_identified = listenable_property.managed(False)
    received_response_bytes = listenable_property.managed(None)

    @depends(send_midi=None)
    def __init__(
        self,
        name="Identification",
        identity_request=midi.SYSEX_IDENTITY_REQUEST_MESSAGE,
        identity_request_delay=0.0,
        identity_response_id_bytes=None,
        custom_identity_response=None,
        send_midi=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._send_midi = send_midi
        self._identity_request = identity_request
        self._responder = create_responder(identity_response_id_bytes, custom_identity_response)
        response_element = self._responder.create_response_element()
        response_element.name = "identity_control"
        response_element.is_private = True
        self.identity_response_control.set_control_element(response_element)
        self._request_task = self._tasks.add(
            task.sequence(
                task.run(self._send_identity_request),
                task.wait(identity_request_delay),
                task.run(self._send_identity_request),
            ),
        )
        self._request_task.kill()

    @identity_response_control.value
    def identity_response_control(self, response_bytes, _):
        try:
            if self._responder.is_valid_response(response_bytes):
                self._request_task.kill()
                self.identity_response_control.enabled = False
                self.received_response_bytes = response_bytes
                self.is_identified = True
                self.notify(self.notifications.identify)
        except IdentityResponseError as e:
            try:
                logger.error(e)
            finally:
                e = None
                del e

    def request_identity(self):
        self._request_task.restart()
        self.received_response_bytes = None
        self.is_identified = False

    def _send_identity_request(self):
        self.identity_response_control.enabled = True
        self._send_midi(self._identity_request)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/notifications/type_decl.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1155 bytes
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union

if TYPE_CHECKING:
    from typing_extensions import ParamSpec, TypeVarTuple

    NotificationParams = TypeVarTuple("NotificationParams")
    NotificationParamSpec = ParamSpec("NotificationParamSpec")
else:
    NotificationParams = TypeVar("NotificationParams")
    NotificationParamSpec = ...


@dataclass
class _TransformDefaultText:
    transform_fn = lambda s: s
    transform_fn: Callable[[str], str]


class _DefaultText:
    pass


NotificationFnType = TypeVar("NotificationFnType", bound=Callable)
Notification = Optional[Union[(str, _DefaultText, _TransformDefaultText, NotificationFnType)]]
Fn = Callable[(NotificationParamSpec, Any)]
NOTIFICATION_EVENT_ID = "notification"



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/TransportComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9732 bytes
from functools import partial

from past.utils import old_div

from . import Task
from .CompoundComponent import CompoundComponent
from .SubjectSlot import subject_slot
from .ToggleComponent import ToggleComponent
from .Util import clamp, const, in_range

TEMPO_TOP = 200.0
TEMPO_BOTTOM = 60.0
TEMPO_FINE_RANGE = 2.56
SEEK_SPEED = 10.0


class TransportComponent(CompoundComponent):
    def __init__(self, play_toggle_model_transform=const(True), *a, **k):
        (super().__init__)(*a, **k)
        self._ffwd_button = None
        self._rwd_button = None
        self._tap_tempo_button = None
        self._tempo_control = None
        self._tempo_fine_control = None
        self._song_position_control = None
        self._rwd_task = Task.Task()
        self._ffwd_task = Task.Task()
        self._fine_tempo_needs_pickup = True
        self._prior_fine_tempo_value = -1
        self._end_undo_step_task = self._tasks.add(Task.sequence(Task.wait(1.5), Task.run(self.song().end_undo_step)))
        self._end_undo_step_task.kill()
        song = self.song()
        (
            self._loop_toggle,
            self._punch_in_toggle,
            self._punch_out_toggle,
            self._record_toggle,
            self._play_toggle,
            self._stop_toggle,
            self._nudge_down_toggle,
            self._nudge_up_toggle,
            self._metronome_toggle,
            self._arrangement_overdub_toggle,
            self._overdub_toggle,
            self._session_record_toggle,
        ) = self.register_components(
            ToggleComponent("loop", song),
            ToggleComponent("punch_in", song, is_momentary=True),
            ToggleComponent("punch_out", song, is_momentary=True),
            ToggleComponent("record_mode", song),
            ToggleComponent("is_playing", song, model_transform=play_toggle_model_transform),
            ToggleComponent("is_playing", song, model_transform=(const(False)), view_transform=(const(False))),
            ToggleComponent("nudge_down", song, is_momentary=True),
            ToggleComponent("nudge_up", song, is_momentary=True),
            ToggleComponent("metronome", song),
            ToggleComponent("arrangement_overdub", song),
            ToggleComponent("overdub", song),
            ToggleComponent("session_record", song),
        )

    def set_stop_button(self, button):
        self._stop_toggle.set_toggle_button(button)

    def set_play_button(self, button):
        self._play_toggle.set_toggle_button(button)

    def set_seek_buttons(self, ffwd_button, rwd_button):
        if self._ffwd_button != ffwd_button:
            self._ffwd_button = ffwd_button
            self._ffwd_value_slot.subject = ffwd_button
            self._ffwd_task.kill()
        if self._rwd_button != rwd_button:
            self._rwd_button = rwd_button
            self._rwd_value_slot.subject = rwd_button
            self._rwd_task.kill()

    def set_seek_forward_button(self, ffwd_button):
        if self._ffwd_button != ffwd_button:
            self._ffwd_button = ffwd_button
            self._ffwd_value_slot.subject = ffwd_button
            self._ffwd_task.kill()

    def set_seek_backward_button(self, rwd_button):
        if self._rwd_button != rwd_button:
            self._rwd_button = rwd_button
            self._rwd_value_slot.subject = rwd_button
            self._rwd_task.kill()

    def set_nudge_buttons(self, up_button, down_button):
        self._nudge_up_toggle.set_toggle_button(up_button)
        self._nudge_down_toggle.set_toggle_button(down_button)

    def set_nudge_up_button(self, up_button):
        self._nudge_up_toggle.set_toggle_button(up_button)

    def set_nudge_down_button(self, down_button):
        self._nudge_down_toggle.set_toggle_button(down_button)

    def set_record_button(self, button):
        self._record_toggle.set_toggle_button(button)

    def set_tap_tempo_button(self, button):
        if self._tap_tempo_button != button:
            self._tap_tempo_button = button
            self._tap_tempo_value.subject = button
            self._update_tap_tempo_button()

    def set_loop_button(self, button):
        self._loop_toggle.set_toggle_button(button)

    def set_punch_buttons(self, in_button, out_button):
        self._punch_in_toggle.set_toggle_button(in_button)
        self._punch_out_toggle.set_toggle_button(out_button)

    def set_punch_in_button(self, in_button):
        self._punch_in_toggle.set_toggle_button(in_button)

    def set_punch_out_button(self, out_button):
        self._punch_out_toggle.set_toggle_button(out_button)

    def set_metronome_button(self, button):
        self._metronome_toggle.set_toggle_button(button)

    def set_arrangement_overdub_button(self, button):
        self._arrangement_overdub_toggle.set_toggle_button(button)

    def set_overdub_button(self, button):
        self._overdub_toggle.set_toggle_button(button)

    def set_session_record_button(self, button):
        self._session_record_toggle.set_toggle_button(button)

    def set_tempo_control(self, control, fine_control=None):
        if self._tempo_control != control:
            self._tempo_control = control
            self._tempo_value.subject = control
        if self._tempo_fine_control != fine_control:
            self._tempo_fine_control = fine_control
            self._tempo_fine_value.subject = fine_control
            self._fine_tempo_needs_pickup = True
            self._prior_fine_tempo_value = -1

    def set_tempo_fine_control(self, fine_control):
        if self._tempo_fine_control != fine_control:
            self._tempo_fine_control = fine_control
            self._tempo_fine_value.subject = fine_control
            self._fine_tempo_needs_pickup = True
            self._prior_fine_tempo_value = -1

    def set_song_position_control(self, control):
        if self._song_position_control != control:
            self._song_position_control = control
            self._song_position_value.subject = control

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_tap_tempo_button()

    @subject_slot("value")
    def _ffwd_value_slot(self, value):
        self._ffwd_value(value)

    def _ffwd_value(self, value):
        if self._ffwd_button.is_momentary():
            self._ffwd_task.kill()
            if value:
                self._ffwd_task = self._tasks.add(partial(self._move_current_song_time, SEEK_SPEED))
        else:
            if self.is_enabled():
                self.song().current_song_time += 1

    @subject_slot("value")
    def _rwd_value_slot(self, value):
        self._rwd_value(value)

    def _rwd_value(self, value):
        if self._rwd_button.is_momentary():
            self._rwd_task.kill()
            if value:
                self._rwd_task = self._tasks.add(partial(self._move_current_song_time, -SEEK_SPEED))
        else:
            if self.is_enabled():
                song = self.song()
                song.current_song_time = max(0.0, song.current_song_time - 1)

    def _move_current_song_time(self, speed, delta):
        song = self.song()
        song.current_song_time = max(0.0, song.current_song_time + speed * delta)
        return Task.RUNNING

    @subject_slot("value")
    def _tap_tempo_value(self, value):
        if self.is_enabled():
            if not (value or self._tap_tempo_button.is_momentary()):
                if not self._end_undo_step_task.is_running:
                    self.song().begin_undo_step()
                self._end_undo_step_task.restart()
                self.song().tap_tempo()
            self._update_tap_tempo_button()

    def _update_tap_tempo_button(self):
        if self.is_enabled():
            if self._tap_tempo_button:
                self._tap_tempo_button.set_light(True)

    @subject_slot("value")
    def _tempo_value(self, value):
        if self.is_enabled():
            fraction = (TEMPO_TOP - TEMPO_BOTTOM) / 127.0
            self.song().tempo = fraction * value + TEMPO_BOTTOM

    @subject_slot("value")
    def _tempo_fine_value(self, value):
        if self.is_enabled():
            if self._fine_tempo_needs_pickup:
                if in_range(self._prior_fine_tempo_value, 0, 128):
                    range_max = max(value, self._prior_fine_tempo_value)
                    range_min = min(value, self._prior_fine_tempo_value)
                    if in_range(64, range_min, range_max + 1):
                        self._fine_tempo_needs_pickup = False
            else:
                difference = value - self._prior_fine_tempo_value
                ratio = 127.0 / TEMPO_FINE_RANGE
                new_tempo = clamp(self.song().tempo + old_div(difference, ratio), TEMPO_BOTTOM, TEMPO_TOP)
                self.song().tempo = new_tempo
        self._prior_fine_tempo_value = value

    @subject_slot("value")
    def _song_position_value(self, value):
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class LomObject:
    """
    this is the base class for an object that is accessible via the LOM
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/item_lister.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 9325 bytes
from ...base import EventObject, forward_property, listenable_property, listens, liveobj_valid
from .. import Component
from ..control import ButtonControl, control_list


class SimpleItemSlot(EventObject):
    def __init__(self, item=None, name="", nesting_level=-1, *a, **k):
        (super().__init__)(*a, **k)
        self._item = item
        self._name = name
        self._nesting_level = nesting_level
        self._SimpleItemSlot__on_name_changed.subject = (
            self._item if getattr(self._item, "name_has_listener", None) else None
        )
        self._SimpleItemSlot__on_color_index_changed.subject = (
            self._item if getattr(self._item, "color_index_has_listener", None) else None
        )

    @listenable_property
    def name(self):
        return self._name

    @property
    def item(self):
        return self._item

    @property
    def nesting_level(self):
        return self._nesting_level

    @listenable_property
    def color_index(self):
        return getattr(self._item, "color_index", -1)

    @listens("name")
    def __on_name_changed(self):
        self.notify_name()
        self._name = self._item.name

    @listens("color_index")
    def __on_color_index_changed(self):
        self.notify_color_index()


class ItemSlot(SimpleItemSlot):
    def __init__(self, item=None, nesting_level=0, **k):
        (super().__init__)(item=item, name=item.name, nesting_level=nesting_level, **k)

    def __eq__(self, other):
        return id(self) == id(other) or self._item == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self._item)

    _live_ptr = forward_property("_item")("_live_ptr")


class ItemProvider(EventObject):
    __events__ = ("items", "selected_item")

    @property
    def items(self):
        return []

    @property
    def selected_item(self):
        pass


class ItemListerComponentBase(Component):
    __events__ = ("items",)

    def __init__(self, item_provider=ItemProvider(), num_visible_items=8, *a, **k):
        (super().__init__)(*a, **k)
        self._item_offset = 0
        self._item_provider = item_provider
        self._items = []
        self._num_visible_items = num_visible_items
        self._ItemListerComponentBase__on_items_changed.subject = item_provider
        self.update_items()

    def reset_offset(self):
        self._item_offset = 0

    @property
    def items(self):
        return self._items

    @property
    def item_provider(self):
        return self._item_provider

    @property
    def item_offset(self):
        return self._item_offset

    @item_offset.setter
    def item_offset(self, offset):
        self._item_offset = offset
        self.update_items()

    def can_scroll_left(self):
        return self.item_offset > 0

    def can_scroll_right(self):
        items = self._item_provider.items[self.item_offset :]
        return len(items) > self._num_visible_items

    def scroll_left(self):
        self.item_offset -= 1

    def scroll_right(self):
        self.item_offset += 1

    def _adjust_offset(self):
        num_raw_items = len(self._item_provider.items)
        list_length = self._num_visible_items
        if list_length >= num_raw_items or self._item_offset >= num_raw_items - list_length:
            self._item_offset = max(0, num_raw_items - list_length)

    def update_items(self):
        for item in self._items:
            self.disconnect_disconnectable(item)

        self._adjust_offset()
        self._items = list(map(self.register_disconnectable, self._create_slots()))
        self.notify_items()

    def _create_slots(self):
        items = self._item_provider.items[self.item_offset :]
        num_slots = min(self._num_visible_items, len(items))
        new_items = []
        if num_slots > 0:
            new_items = [
                (self._create_slot)(index, *item)
                for index, item in enumerate(items[:num_slots])
                if liveobj_valid(item[0])
            ]
        return new_items

    def _create_slot(self, index, item, nesting_level):
        return ItemSlot(item=item, nesting_level=nesting_level)

    @listens("items")
    def __on_items_changed(self):
        self.update_items()
        self._on_items_changed()

    def _on_items_changed(self):
        pass


class ScrollComponent(Component):
    __events__ = ("scroll",)
    button = ButtonControl(color="ItemNavigation.ItemNotSelected", repeat=True)

    @button.pressed
    def button(self, button):
        self.notify_scroll()


class ScrollOverlayComponent(Component):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._scroll_left_component, self._scroll_right_component = self.add_children(
            ScrollComponent(is_enabled=False),
            ScrollComponent(is_enabled=False),
        )
        self._ScrollOverlayComponent__on_scroll_left.subject = self._scroll_left_component
        self._ScrollOverlayComponent__on_scroll_right.subject = self._scroll_right_component

    scroll_left_layer = forward_property("_scroll_left_component")("layer")
    scroll_right_layer = forward_property("_scroll_right_component")("layer")

    def can_scroll_left(self):
        raise NotImplementedError

    def can_scroll_right(self):
        raise NotImplementedError

    def scroll_left(self):
        raise NotImplementedError

    def scroll_right(self):
        raise NotImplementedError

    def update_scroll_buttons(self):
        if self.is_enabled():
            self._scroll_left_component.set_enabled(self.can_scroll_left())
            self._scroll_right_component.set_enabled(self.can_scroll_right())

    @listens("scroll")
    def __on_scroll_left(self):
        self.scroll_left()

    @listens("scroll")
    def __on_scroll_right(self):
        self.scroll_right()

    def update(self):
        super().update()
        if self.is_enabled():
            self.update_scroll_buttons()


class ItemListerComponent(ItemListerComponentBase):
    color_class_name = "ItemNavigation"
    select_buttons = control_list(ButtonControl, unavailable_color=(color_class_name + ".NoItem"))

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._scroll_overlay = self.add_children(ScrollOverlayComponent(is_enabled=True))
        self._scroll_overlay.can_scroll_left = self.can_scroll_left
        self._scroll_overlay.can_scroll_right = self.can_scroll_right
        self._scroll_overlay.scroll_left = self.scroll_left
        self._scroll_overlay.scroll_right = self.scroll_right
        self._ItemListerComponent__on_items_changed.subject = self
        self._ItemListerComponent__on_selection_changed.subject = self._item_provider

    scroll_left_layer = forward_property("_scroll_overlay")("scroll_left_layer")
    scroll_right_layer = forward_property("_scroll_overlay")("scroll_right_layer")

    @listens("items")
    def __on_items_changed(self):
        self.select_buttons.control_count = len(self.items)
        self._update_button_colors()
        self._scroll_overlay.update_scroll_buttons()

    @listens("selected_item")
    def __on_selection_changed(self):
        self._on_selection_changed()

    def _on_selection_changed(self):
        self._update_button_colors()

    def _items_equal(self, item, selected_item):
        return item == selected_item

    def _update_button_colors(self):
        selected_item = self._item_provider.selected_item
        for button, item in zip(self.select_buttons, self.items):
            button.color = self._color_for_button(button.index, self._items_equal(item, selected_item))

    def _color_for_button(self, button_index, is_selected):
        if is_selected:
            return self.color_class_name + ".ItemSelected"
        return self.color_class_name + ".ItemNotSelected"

    @select_buttons.pressed
    def select_buttons(self, button):
        self._on_select_button_pressed(button)

    @select_buttons.pressed_delayed
    def select_buttons(self, button):
        self._on_select_button_pressed_delayed(button)

    @select_buttons.released
    def select_buttons(self, button):
        self._on_select_button_released(button)

    @select_buttons.released_immediately
    def select_buttons(self, button):
        self._on_select_button_released_immediately(button)

    def _on_select_button_pressed(self, button):
        pass

    def _on_select_button_pressed_delayed(self, button):
        pass

    def _on_select_button_released(self, button):
        pass

    def _on_select_button_released_immediately(self, button):
        pass
